# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import bz2
import collections
from datetime import timedelta
from itertools import dropwhile, islice
import lzma
from typing import (
    Callable,
    Dict,
    Iterable,
    Iterator,
    Literal,
    Mapping,
    Optional,
    Tuple,
    Union,
)
import zlib

from typing_extensions import Protocol

from swh.model import hashutil
from swh.model.model import Sha1
from swh.objstorage.constants import DEFAULT_LIMIT, ID_HASH_ALGO
from swh.objstorage.exc import ObjCorruptedError, ObjNotFoundError
from swh.objstorage.interface import CompositeObjId, ObjId, ObjStorageInterface


def objid_to_default_hex(
    obj_id: ObjId, algo: Literal["sha1", "sha256"] = ID_HASH_ALGO
) -> str:
    """Converts SHA1 hashes and multi-hashes to the hexadecimal representation
    of the SHA1."""
    if isinstance(obj_id, bytes):
        return hashutil.hash_to_hex(obj_id)
    elif isinstance(obj_id, str):
        return obj_id
    else:
        return hashutil.hash_to_hex(obj_id[algo])


def compute_hashes(
    content: bytes, hash_names: Iterable[str] = hashutil.DEFAULT_ALGORITHMS
) -> Dict[str, bytes]:
    """Compute the content's hashes.

    Args:
        content: The raw content to hash
        hash_names: Names of hashing algorithms
            (default to :const:`swh.model.hashutil.DEFAULT_ALGORITHMS`)

    Returns:
        A dict mapping algo name to hash value

    """
    return hashutil.MultiHash.from_data(
        content,
        hash_names=hash_names,
    ).digest()


def compute_hash(content: bytes, algo: str = ID_HASH_ALGO) -> bytes:
    """Compute the content's hash.

    Args:
        content: The raw content to hash
        hash_name: Hash's name

    Returns:
        The computed hash for the content

    """
    return compute_hashes(content, [algo])[algo]


class NullCompressor:
    def compress(self, data):
        return data

    def flush(self):
        return b""


class NullDecompressor:
    def decompress(self, data: bytes) -> bytes:
        return data

    @property
    def unused_data(self) -> bytes:
        return b""


class _CompressorProtocol(Protocol):
    def compress(self, data: bytes) -> bytes:
        ...

    def flush(self) -> bytes:
        ...


class _DecompressorProtocol(Protocol):
    def decompress(self, data: bytes) -> bytes:
        ...

    unused_data: bytes


decompressors: Dict[str, Callable[[], _DecompressorProtocol]] = {
    "bz2": bz2.BZ2Decompressor,  # type: ignore
    "lzma": lzma.LZMADecompressor,  # type: ignore
    "gzip": lambda: zlib.decompressobj(wbits=31),
    "zlib": zlib.decompressobj,
    "none": NullDecompressor,  # type: ignore
}

compressors: Dict[str, Callable[[], _CompressorProtocol]] = {
    "bz2": bz2.BZ2Compressor,
    "lzma": lzma.LZMACompressor,
    "gzip": lambda: zlib.compressobj(wbits=31),
    "zlib": zlib.compressobj,
    "none": NullCompressor,
}

CompressionFormat = Literal["bz2", "lzma", "gzip", "zlib", "none"]


class ObjStorage(metaclass=abc.ABCMeta):
    PRIMARY_HASH: Literal["sha1", "sha256"] = "sha1"
    compression: CompressionFormat = "none"

    def __init__(self, *, allow_delete=False, **kwargs):
        # A more complete permission system could be used in place of that if
        # it becomes needed
        self.allow_delete = allow_delete

    def add_batch(
        self: ObjStorageInterface,
        contents: Union[Mapping[Sha1, bytes], Iterable[Tuple[ObjId, bytes]]],
        check_presence: bool = True,
    ) -> Dict:
        summary = {"object:add": 0, "object:add:bytes": 0}
        contents_pairs: Iterable[Tuple[ObjId, bytes]]
        if isinstance(contents, collections.abc.Mapping):
            contents_pairs = contents.items()
        else:
            contents_pairs = contents
        for obj_id, content in contents_pairs:
            if check_presence and obj_id in self:
                continue
            self.add(content, obj_id, check_presence=False)
            summary["object:add"] += 1
            summary["object:add:bytes"] += len(content)
        return summary

    def restore(self: ObjStorageInterface, content: bytes, obj_id: ObjId) -> None:
        # check_presence to false will erase the potential previous content.
        self.add(content, obj_id, check_presence=False)

    def get_batch(
        self: ObjStorageInterface, obj_ids: Iterable[ObjId]
    ) -> Iterator[Optional[bytes]]:
        for obj_id in obj_ids:
            try:
                yield self.get(obj_id)
            except ObjNotFoundError:
                yield None

    @abc.abstractmethod
    def delete(self, obj_id: ObjId):
        if not self.allow_delete:
            raise PermissionError("Delete is not allowed.")

    def list_content(
        self: ObjStorageInterface,
        last_obj_id: Optional[ObjId] = None,
        limit: Optional[int] = DEFAULT_LIMIT,
    ) -> Iterator[CompositeObjId]:
        it = iter(self)
        if last_obj_id:
            last_obj_id_hex = objid_to_default_hex(last_obj_id)
            it = dropwhile(lambda x: objid_to_default_hex(x) <= last_obj_id_hex, it)
        return islice(it, limit)

    def download_url(
        self,
        obj_id: ObjId,
        content_disposition: Optional[str] = None,
        expiry: Optional[timedelta] = None,
    ) -> Optional[str]:
        return None

    @abc.abstractmethod
    def get(self, obj_id: ObjId) -> bytes:
        raise NotImplementedError()

    def check(self, obj_id: ObjId) -> None:
        """Check if a content is found and recompute its hash to check integrity."""
        obj_content = self.get(obj_id)
        hash_algos = [str(self.PRIMARY_HASH)]
        if isinstance(obj_id, dict):
            hash_algos += [algo for algo in obj_id if algo != self.PRIMARY_HASH]
        actual_hashes = compute_hashes(obj_content, hash_algos)
        for algo in hash_algos:
            actual_obj_id = actual_hashes[algo]
            expected_obj_id = obj_id
            if isinstance(obj_id, dict):
                expected_obj_id = obj_id[algo]  # type: ignore[literal-required]
            if actual_obj_id != expected_obj_id:
                raise ObjCorruptedError(
                    f"expected {algo} hash is {hashutil.hash_to_hex(expected_obj_id)}, "
                    f"actual {algo} hash is {hashutil.hash_to_hex(actual_obj_id)}"
                )

    def compress(self, data: bytes) -> bytes:
        compressor = compressors[self.compression]()
        compressed = compressor.compress(data)
        compressed += compressor.flush()
        return compressed

    def decompress(self, data: bytes, hex_obj_id: str) -> bytes:
        decompressor = decompressors[self.compression]()
        try:
            ret = decompressor.decompress(data)
        except (zlib.error, lzma.LZMAError, OSError):
            raise ObjCorruptedError(
                f"content with {self.PRIMARY_HASH} hash {hex_obj_id} is not a proper "
                "compressed file"
            )
        if decompressor.unused_data:
            raise ObjCorruptedError(
                f"trailing data found when decompressing content with {self.PRIMARY_HASH} "
                f"{hex_obj_id}"
            )
        return ret



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from functools import partial
import logging
from multiprocessing import Process
from typing import Callable, List, Optional, Tuple

from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.interface import ObjId
from swh.objstorage.objstorage import ObjStorage

from .roshard import (
    DEFAULT_IMAGE_FEATURES_UNSUPPORTED,
    Pool,
    ROShard,
    ROShardCreator,
    ShardNotMapped,
)
from .rwshard import RWShard
from .sharedbase import ShardState, SharedBase
from .sleep import sleep_exponential
from .stats import Stats

logger = logging.getLogger(__name__)


class WineryObjStorage(ObjStorage):
    PRIMARY_HASH = "sha256"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if kwargs.get("readonly"):
            self.winery = WineryReader(**kwargs)
        else:
            self.winery = WineryWriter(**kwargs)

    def get(self, obj_id: ObjId) -> bytes:
        return self.winery.get(self._hash(obj_id))

    def check_config(self, *, check_write: bool) -> bool:
        return True

    def __contains__(self, obj_id: ObjId) -> bool:
        return self._hash(obj_id) in self.winery

    def add(self, content: bytes, obj_id: ObjId, check_presence: bool = True) -> None:
        self.winery.add(content, self._hash(obj_id), check_presence)

    def delete(self, obj_id: ObjId):
        if not self.allow_delete:
            raise PermissionError("Delete is not allowed.")
        return self.winery.delete(obj_id)

    def _hash(self, obj_id: ObjId) -> bytes:
        if isinstance(obj_id, dict):
            return obj_id[self.PRIMARY_HASH]
        else:
            return obj_id

    def on_shutdown(self):
        self.winery.on_shutdown()


class WineryBase:
    def __init__(self, **kwargs):
        self.args = kwargs
        self.base = SharedBase(**self.args)

    def __contains__(self, obj_id):
        return self.base.contains(obj_id)

    def on_shutdown(self):
        return


class WineryReader(WineryBase):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.ro_shards = {}
        self.rw_shards = {}

    def roshard(self, name) -> Optional[ROShard]:
        if name not in self.ro_shards:
            try:
                shard = ROShard(name, **self.args)
            except ShardNotMapped:
                return None
            self.ro_shards[name] = shard
            if name in self.rw_shards:
                del self.rw_shards[name]
        return self.ro_shards[name]

    def rwshard(self, name) -> RWShard:
        if name not in self.rw_shards:
            shard = RWShard(name, **self.args)
            self.rw_shards[name] = shard
        return self.rw_shards[name]

    def get(self, obj_id: bytes) -> bytes:
        shard_info = self.base.get(obj_id)
        if shard_info is None:
            raise ObjNotFoundError(obj_id)
        name, state = shard_info
        content: Optional[bytes] = None
        if state.image_available:
            roshard = self.roshard(name)
            if roshard:
                content = roshard.get(obj_id)
        if content is None:
            rwshard = self.rwshard(name)
            content = rwshard.get(obj_id)
        if content is None:
            raise ObjNotFoundError(obj_id)
        return content


def pack(shard, shared_base=None, clean_immediately=False, **kwargs) -> bool:
    stats = Stats(kwargs.get("output_dir"))
    rw = RWShard(shard, **kwargs)

    count = rw.count()
    logger.info("Creating RO shard %s for %s objects", shard, count)
    with ROShardCreator(shard, count, **kwargs) as ro:
        logger.info("Created RO shard %s", shard)
        for i, (obj_id, content) in enumerate(rw.all()):
            ro.add(content, obj_id)
            if stats.stats_active:
                stats.stats_read(obj_id, content)
                stats.stats_write(obj_id, content)
            if i % 100 == 99:
                logger.debug("RO shard %s: added %s/%s objects", shard, i + 1, count)

        logger.debug("RO shard %s: added %s objects, saving", shard, count)

    logger.info("RO shard %s: saved", shard)

    if not shared_base:
        shared_base = SharedBase(**kwargs)
    shared_base.shard_packing_ends(shard)
    if clean_immediately:
        cleanup_rw_shard(shard, shared_base=shared_base, **kwargs)
    return True


def cleanup_rw_shard(shard, shared_base=None, **kwargs) -> bool:
    rw = RWShard(shard, **{"shard_max_size": 0, **kwargs})

    rw.drop()

    if not shared_base:
        shared_base = SharedBase(**kwargs)
    shared_base.set_shard_state(name=shard, new_state=ShardState.READONLY)

    return True


class WineryWriter(WineryReader):
    def __init__(
        self,
        pack_immediately: bool = True,
        clean_immediately: bool = True,
        rwshard_idle_timeout: float = 300,
        **kwargs,
    ):
        self.pack_immediately = pack_immediately
        self.clean_immediately = clean_immediately
        super().__init__(**kwargs)
        self.shards_filled: List[str] = []
        self.packers: List[Process] = []
        self._shard: Optional[RWShard] = None
        self.idle_timeout = rwshard_idle_timeout

    def release_shard(
        self,
        shard: Optional[RWShard] = None,
        from_idle_handler: bool = False,
        new_state: ShardState = ShardState.STANDBY,
    ):
        """Release the currently locked shard"""
        if not shard:
            shard = self._shard

        if not shard:
            return

        logger.debug("WineryWriter releasing shard %s", shard.name)

        self.base.set_shard_state(new_state=new_state, name=shard.name)
        if not from_idle_handler:
            logger.debug("Shard released, disabling idle handler")
            shard.disable_idle_handler()
        self._shard = None

    @property
    def shard(self):
        """Lock a shard to be able to use it. Release it after :attr:`idle_timeout`."""
        if not self._shard:
            self._shard = RWShard(
                self.base.locked_shard,
                idle_timeout_cb=partial(self.release_shard, from_idle_handler=True),
                idle_timeout=self.idle_timeout,
                **self.args,
            )
            logger.debug(
                "WineryBase: locked RWShard %s, releasing it in %s",
                self._shard.name,
                self.idle_timeout,
            )
        return self._shard

    def add(self, content: bytes, obj_id: bytes, check_presence: bool = True) -> None:
        if check_presence and obj_id in self:
            return

        with self.base.pool.connection() as db, db.transaction():
            shard = self.base.record_new_obj_id(db, obj_id)
            if shard != self.base.locked_shard_id:
                #  this object is the responsibility of another shard
                return

            self.shard.add(db, obj_id, content)

        if self.shard.is_full():
            filled_name = self.shard.name
            self.release_shard(new_state=ShardState.FULL)
            self.shards_filled.append(filled_name)
            if self.pack_immediately:
                self.pack(filled_name)

    def delete(self, obj_id: bytes):
        shard_info = self.base.get(obj_id)
        if shard_info is None:
            raise ObjNotFoundError(obj_id)
        name, state = shard_info
        # We only care about RWShard for now. ROShards will be
        # taken care in a batch job.
        if not state.image_available:
            rwshard = self.rwshard(name)
            try:
                rwshard.delete(obj_id)
            except KeyError:
                logger.warning(
                    "Shard %s does not seem to know about object %s, but we "
                    "had an entry in SharedBase (which is going to "
                    "be removed just now)",
                    rwshard.name,
                    obj_id,
                )
        self.base.delete(obj_id)

    def check(self, obj_id: ObjId) -> None:
        # load all shards packing == True and not locked (i.e. packer
        # was interrupted for whatever reason) run pack for each of them
        pass

    def pack(self, shard_name: str):
        self.base.shard_packing_starts(shard_name)
        p = Process(
            target=pack,
            kwargs={
                "shard": shard_name,
                "clean_immediately": self.clean_immediately,
                **self.args,
            },
        )
        p.start()
        self.packers.append(p)

    def on_shutdown(self):
        self.release_shard()

    def __del__(self):
        for p in getattr(self, "packers", []):
            if not p.is_alive():
                continue
            logger.warning("Killing packer %s", p)
            p.kill()
            p.join()


def never_stop(_: int) -> bool:
    return False


def stop_after_shards(max_shards_packed: int) -> Callable[[int], bool]:
    def stop(shards_packed: int):
        return shards_packed >= max_shards_packed

    return stop


def shard_packer(
    base_dsn: str,
    shard_max_size: int,
    throttle_read: int,
    throttle_write: int,
    application_name: Optional[str] = None,
    rbd_pool_name: str = "shards",
    rbd_data_pool_name: Optional[str] = None,
    rbd_image_features_unsupported: Tuple[
        str, ...
    ] = DEFAULT_IMAGE_FEATURES_UNSUPPORTED,
    rbd_use_sudo: bool = True,
    rbd_map_options: str = "",
    rbd_create_images: bool = True,
    rbd_wait_for_image: Callable[[int], None] = sleep_exponential(
        min_duration=5,
        factor=2,
        max_duration=60,
        message="Waiting for RBD image mapping",
    ),
    output_dir: Optional[str] = None,
    stop_packing: Callable[[int], bool] = never_stop,
    wait_for_shard: Callable[[int], None] = sleep_exponential(
        min_duration=5,
        factor=2,
        max_duration=60,
        message="No shards to pack",
    ),
) -> int:
    """Pack shards until the `stop_packing` function returns True.

    When no shards are available for packing, call the `wait_for_shard` function.

    Arguments:
      base_dsn: PostgreSQL dsn for the shared database
      shard_max_size: Max size of a shard (used to size new shards)
      throttle_read: reads per second
      throttle_write: writes per second
      application_name: the application name sent to PostgreSQL
      rbd_create_images: create images directly (or wait for RBD mapper)
      rbd_wait_for_image: sleep function called to wait for an image (when
       `rbd_create_images`=`False`)
      rbd_*: passed directly to :class:`roshard.Pool`
      output_dir: output directory for statistics
      stop_packing: callback to determine whether the packer should exit
      wait_for_shard: sleep function called when no shards are available to be packed
    """
    application_name = application_name or "Winery Shard Packer"

    base = SharedBase(base_dsn=base_dsn, application_name=application_name)

    shards_packed = 0
    waited_for_shards = 0
    while not stop_packing(shards_packed):
        locked = base.maybe_lock_one_shard(
            current_state=ShardState.FULL, new_state=ShardState.PACKING
        )

        if not locked:
            wait_for_shard(waited_for_shards)
            waited_for_shards += 1
            continue

        waited_for_shards = 0

        with locked:
            logger.info("shard_packer: Locked shard %s to pack", locked.name)
            ret = pack(
                locked.name,
                base_dsn=base_dsn,
                shard_max_size=shard_max_size,
                output_dir=output_dir,
                shared_base=base,
                throttle_read=throttle_read,
                throttle_write=throttle_write,
                application_name=application_name,
                rbd_use_sudo=rbd_use_sudo,
                rbd_map_options=rbd_map_options,
                rbd_create_images=rbd_create_images,
                rbd_wait_for_image=rbd_wait_for_image,
                rbd_pool_name=rbd_pool_name,
                rbd_data_pool_name=rbd_data_pool_name,
                rbd_image_features_unsupported=rbd_image_features_unsupported,
            )
            if not ret:
                raise ValueError("Packing shard %s failed" % locked.name)
            shards_packed += 1

    return shards_packed


def rw_shard_cleaner(
    base_dsn: str,
    min_mapped_hosts: int,
    application_name: Optional[str] = None,
    stop_cleaning: Callable[[int], bool] = never_stop,
    wait_for_shard: Callable[[int], None] = sleep_exponential(
        min_duration=5,
        factor=2,
        max_duration=60,
        message="No shards to clean up",
    ),
) -> int:
    """Clean up RW shards until the `stop_cleaning` function returns True.

    When no shards are available for packing, call the `wait_for_shard` function.

    Arguments:
      base_dsn: PostgreSQL dsn for the shared database
      min_mapped_hosts: how many hosts should have mapped the image read-only before
        cleaning it
      application_name: the application name sent to PostgreSQL
      stop_cleaning: callback to determine whether the cleaner should exit
      wait_for_shard: sleep function called when no shards are available to be cleaned
    """
    application_name = application_name or "Winery RW shard cleaner"
    base = SharedBase(base_dsn=base_dsn, application_name=application_name)

    shards_cleaned = 0
    waited_for_shards = 0
    while not stop_cleaning(shards_cleaned):
        locked = base.maybe_lock_one_shard(
            current_state=ShardState.PACKED,
            new_state=ShardState.CLEANING,
            min_mapped_hosts=min_mapped_hosts,
        )

        if not locked:
            wait_for_shard(waited_for_shards)
            waited_for_shards += 1
            continue

        waited_for_shards = 0

        with locked:
            logger.info("rw_shard_cleaner: Locked shard %s to clean", locked.name)

            ret = cleanup_rw_shard(
                locked.name,
                base_dsn=base_dsn,
                shared_base=base,
                application_name=application_name,
            )
            if not ret:
                raise ValueError("Cleaning shard %s failed" % locked.name)

            shards_cleaned += 1

    return shards_cleaned


def deleted_objects_cleaner(
    base: SharedBase,
    pool: Pool,
    stop_running: Callable[[], bool],
):
    """Clean up deleted objects from RO shards and the shared database.

    This requires the ability to map RBD images in read-write mode. Images will be
    left mapped by this process as it is meant to be executed in a transient host
    dedicated to this purpose.

    Arguments:
      base_dsn: PostgreSQL dsn for the shared database
      pool: Ceph RBD pool for Winery shards
      stop_running: callback that returns True when the manager should stop running
    """
    count = 0
    for obj_id, shard_name, shard_state in base.deleted_objects():
        if stop_running():
            break
        if shard_state.readonly:
            ROShard.delete(pool, shard_name, obj_id)
        base.clean_deleted_object(obj_id)
        count += 1

    logger.info("Cleaned %d deleted objects", count)



# Copyright (C) 2015-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import os
import time
from types import FrameType
from typing import Optional

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

logger = logging.getLogger(__name__)


@swh_cli_group.group(name="objstorage", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(
        exists=True,
        dir_okay=False,
    ),
    help="Configuration file.",
)
@click.pass_context
def objstorage_cli_group(ctx, config_file):
    """Software Heritage Objstorage tools."""
    from swh.core import config

    if not config_file:
        config_file = os.environ.get("SWH_CONFIG_FILENAME")

    if config_file:
        if not os.path.exists(config_file):
            raise click.ClickException(
                "Configuration file %s does not exist" % config_file
            )
        conf = config.read(config_file)
    else:
        conf = {}

    ctx.ensure_object(dict)

    ctx.obj["config"] = conf


# for BW compat
cli = objstorage_cli_group


@objstorage_cli_group.command("rpc-serve")
@click.option(
    "--host",
    default="0.0.0.0",
    metavar="IP",
    show_default=True,
    help="Host ip address to bind the server on",
)
@click.option(
    "--port",
    "-p",
    default=5003,
    type=click.INT,
    metavar="PORT",
    show_default=True,
    help="Binding port of the server",
)
@click.option(
    "--debug/--no-debug",
    default=True,
    help="Indicates if the server should run in debug mode",
)
@click.pass_context
def serve(ctx, host, port, debug):
    """Run a standalone objstorage server.

    This is not meant to be run on production systems.
    """
    from swh.objstorage.api.server import app, validate_config

    if "log_level" in ctx.obj:
        logging.getLogger("werkzeug").setLevel(ctx.obj["log_level"])
    validate_config(ctx.obj["config"])
    app.config.update(ctx.obj["config"])
    app.run(host, port=int(port), debug=bool(debug))


@objstorage_cli_group.group("winery")
@click.pass_context
def winery(ctx):
    "Winery related commands"
    config = ctx.obj["config"]["objstorage"]
    if config["cls"] != "winery":
        raise click.ClickException("winery packer only works on a winery objstorage")


@winery.command("packer")
@click.option("--stop-after-shards", type=click.INT, default=None)
@click.pass_context
def winery_packer(ctx, stop_after_shards: Optional[int] = None):
    """Run a winery packer process"""
    import signal

    from swh.objstorage.backends.winery.objstorage import shard_packer
    from swh.objstorage.backends.winery.roshard import (
        DEFAULT_IMAGE_FEATURES_UNSUPPORTED,
    )

    config = ctx.obj["config"]["objstorage"]

    signal_received = False

    def stop_packing(num_shards: int) -> bool:
        """Stop packing when a signal is received, or when stop_after_shards is reached."""
        return signal_received or (
            stop_after_shards is not None and num_shards >= stop_after_shards
        )

    def set_signal_received(signum: int, _stack_frame: Optional[FrameType]) -> None:
        nonlocal signal_received
        logger.warning("Received signal %s, exiting", signal.strsignal(signum))
        signal_received = True

    base_dsn = config["base_dsn"]
    shard_max_size = config["shard_max_size"]
    throttle_read = config.get("throttle_read", 200 * 1024 * 1024)
    throttle_write = config.get("throttle_write", 200 * 1024 * 1024)
    output_dir = config.get("output_dir")
    rbd_pool_name = config.get("rbd_pool_name", "shards")
    rbd_data_pool_name = config.get("rbd_data_pool_name")
    rbd_use_sudo = config.get("rbd_use_sudo", True)
    rbd_image_features_unsupported = tuple(
        config.get("rbd_image_features_unsupported", DEFAULT_IMAGE_FEATURES_UNSUPPORTED)
    )
    rbd_create_images = config.get("rbd_create_images", True)
    rbd_map_options = config.get("rbd_map_options", "")

    signal.signal(signal.SIGINT, set_signal_received)
    signal.signal(signal.SIGTERM, set_signal_received)

    ret = shard_packer(
        base_dsn=base_dsn,
        shard_max_size=shard_max_size,
        throttle_read=throttle_read,
        throttle_write=throttle_write,
        rbd_pool_name=rbd_pool_name,
        rbd_data_pool_name=rbd_data_pool_name,
        rbd_image_features_unsupported=rbd_image_features_unsupported,
        rbd_use_sudo=rbd_use_sudo,
        rbd_map_options=rbd_map_options,
        rbd_create_images=rbd_create_images,
        output_dir=output_dir,
        stop_packing=stop_packing,
    )

    logger.info("Packed %s shards", ret)


@winery.command("rbd")
@click.option("--stop-instead-of-waiting", is_flag=True)
@click.pass_context
def winery_rbd(ctx, stop_instead_of_waiting: bool = False):
    """Run a winery RBD image manager process"""
    import signal

    from swh.objstorage.backends.winery.roshard import (
        DEFAULT_IMAGE_FEATURES_UNSUPPORTED,
        Pool,
    )
    from swh.objstorage.backends.winery.sleep import sleep_exponential

    config = ctx.obj["config"]["objstorage"]

    stop_on_next_iteration = False

    def stop_running() -> bool:
        """Stop running when a signal is received, or when there's nothing to do."""
        return stop_on_next_iteration

    def wait_for_image(attempt: int):
        nonlocal stop_on_next_iteration
        if stop_instead_of_waiting:
            stop_on_next_iteration = True
            return

        return sleep_exponential(
            min_duration=1,
            max_duration=60,
            factor=2,
            message="No new RBD images",
        )(attempt)

    def set_signal_received(signum: int, _stack_frame: Optional[FrameType]) -> None:
        nonlocal stop_on_next_iteration
        logger.warning("Received signal %s, exiting", signal.strsignal(signum))
        stop_on_next_iteration = True

    base_dsn = config["base_dsn"]
    shard_max_size = config["shard_max_size"]
    rbd_pool_name = config.get("rbd_pool_name", "shards")
    rbd_data_pool_name = config.get("rbd_data_pool_name")
    rbd_use_sudo = config.get("rbd_use_sudo", True)
    rbd_image_features_unsupported = tuple(
        config.get("rbd_image_features_unsupported", DEFAULT_IMAGE_FEATURES_UNSUPPORTED)
    )
    rbd_manage_rw_images = config.get("rbd_manage_rw_images", True)
    rbd_map_options = config.get("rbd_map_options", "")

    signal.signal(signal.SIGINT, set_signal_received)
    signal.signal(signal.SIGTERM, set_signal_received)

    pool = Pool(
        shard_max_size=shard_max_size,
        rbd_pool_name=rbd_pool_name,
        rbd_data_pool_name=rbd_data_pool_name,
        rbd_use_sudo=rbd_use_sudo,
        rbd_image_features_unsupported=rbd_image_features_unsupported,
        rbd_map_options=rbd_map_options,
    )

    pool.manage_images(
        base_dsn=base_dsn,
        manage_rw_images=rbd_manage_rw_images,
        wait_for_image=wait_for_image,
        stop_running=stop_running,
    )

    logger.info("Image manager exiting")


@winery.command("rw-shard-cleaner")
@click.option("--stop-after-shards", type=click.INT, default=None)
@click.option("--stop-instead-of-waiting", is_flag=True)
@click.option(
    "--min-mapped-hosts",
    type=click.INT,
    default=1,
    help="Number of hosts on which the image should be mapped read-only before cleanup",
)
@click.pass_context
def winery_rw_shard_cleaner(
    ctx,
    stop_after_shards: Optional[int] = None,
    stop_instead_of_waiting: bool = False,
    min_mapped_hosts: int = 1,
):
    """Run a winery RBD image manager process"""
    import signal

    from swh.objstorage.backends.winery.objstorage import rw_shard_cleaner
    from swh.objstorage.backends.winery.sleep import sleep_exponential

    config = ctx.obj["config"]["objstorage"]

    stop_on_next_iteration = False

    def stop_cleaning(num_shards: int) -> bool:
        """Stop running when requested, or when the max number of shards was reached."""
        return (
            stop_after_shards is not None and num_shards >= stop_after_shards
        ) or stop_on_next_iteration

    def wait_for_shard(attempt: int):
        nonlocal stop_on_next_iteration
        if stop_instead_of_waiting:
            stop_on_next_iteration = True
            return

        return sleep_exponential(
            min_duration=1,
            max_duration=60,
            factor=2,
            message="No shards to clean up",
        )(attempt)

    def set_signal_received(signum: int, _stack_frame: Optional[FrameType]) -> None:
        nonlocal stop_on_next_iteration
        logger.warning("Received signal %s, exiting", signal.strsignal(signum))
        stop_on_next_iteration = True

    base_dsn = config["base_dsn"]

    signal.signal(signal.SIGINT, set_signal_received)
    signal.signal(signal.SIGTERM, set_signal_received)

    ret = rw_shard_cleaner(
        base_dsn=base_dsn,
        min_mapped_hosts=min_mapped_hosts,
        stop_cleaning=stop_cleaning,
        wait_for_shard=wait_for_shard,
    )

    logger.info("RW shard cleaner exiting, %d shards cleaned", ret)


@winery.command("clean-deleted-objects")
@click.pass_context
def winery_clean_deleted_objects(ctx):
    """Clean deleted objects from Winery"""
    import signal

    from swh.objstorage.backends.winery.objstorage import deleted_objects_cleaner
    from swh.objstorage.backends.winery.roshard import (
        DEFAULT_IMAGE_FEATURES_UNSUPPORTED,
        Pool,
    )
    from swh.objstorage.backends.winery.sharedbase import SharedBase

    config = ctx.obj["config"]["objstorage"]
    base_dsn = config["base_dsn"]
    shard_max_size = config["shard_max_size"]
    rbd_pool_name = config.get("rbd_pool_name", "shards")
    rbd_data_pool_name = config.get("rbd_data_pool_name")
    rbd_use_sudo = config.get("rbd_use_sudo", True)
    rbd_image_features_unsupported = tuple(
        config.get("rbd_image_features_unsupported", DEFAULT_IMAGE_FEATURES_UNSUPPORTED)
    )

    stop_on_next_iteration = False

    def stop_running() -> bool:
        """Stop running when a signal is received, or when there's nothing to do."""
        return stop_on_next_iteration

    def set_signal_received(signum: int, _stack_frame: Optional[FrameType]) -> None:
        nonlocal stop_on_next_iteration
        logger.warning("Received signal %s, exiting", signal.strsignal(signum))
        stop_on_next_iteration = True

    signal.signal(signal.SIGINT, set_signal_received)
    signal.signal(signal.SIGTERM, set_signal_received)

    base = SharedBase(base_dsn=base_dsn)

    pool = Pool(
        shard_max_size=shard_max_size,
        rbd_pool_name=rbd_pool_name,
        rbd_data_pool_name=rbd_data_pool_name,
        rbd_use_sudo=rbd_use_sudo,
        rbd_image_features_unsupported=rbd_image_features_unsupported,
    )

    deleted_objects_cleaner(base, pool, stop_running)


@objstorage_cli_group.command("import")
@click.argument("directory", required=True, nargs=-1)
@click.pass_context
def import_directories(ctx, directory):
    """Import a local directory in an existing objstorage."""
    from swh.objstorage.factory import get_objstorage
    from swh.objstorage.objstorage import compute_hash

    objstorage = get_objstorage(**ctx.obj["config"]["objstorage"])
    nobj = 0
    volume = 0
    t0 = time.time()
    for dirname in directory:
        for root, _dirs, files in os.walk(dirname):
            for name in files:
                path = os.path.join(root, name)
                with open(path, "rb") as f:
                    content = f.read()
                objstorage.add(content, obj_id=compute_hash(content))
                volume += os.stat(path).st_size
                nobj += 1
    click.echo(
        "Imported %d files for a volume of %s bytes in %d seconds"
        % (nobj, volume, time.time() - t0)
    )


@objstorage_cli_group.command("fsck")
@click.pass_context
def fsck(ctx):
    """Check the objstorage is not corrupted."""
    from swh.objstorage.factory import get_objstorage

    objstorage = get_objstorage(**ctx.obj["config"]["objstorage"])
    for obj_id in objstorage:
        try:
            objstorage.check(obj_id)
        except objstorage.Error as err:
            logging.error(err)


def main():
    return cli(auto_envvar_prefix="SWH_OBJSTORAGE")


if __name__ == "__main__":
    main()



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import time
from typing import Callable

logger = logging.getLogger(__name__)


def sleep_exponential(
    min_duration: float, factor: float, max_duration: float, message: str
) -> Callable[[int], None]:
    """Return a function that returns a callback that sleeps `min_duration`,
    then increases that by `factor` at every call, up to `max_duration`."""
    if min(min_duration, max_duration) <= 0:
        raise ValueError("Cannot sleep for a negative amount of time")

    def sleep(attempt: int):
        duration = min(max_duration, min_duration * factor**attempt)
        logger.debug("%s. Waiting for %s", message, duration)
        time.sleep(duration)

    return sleep



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
import pytest

from swh.objstorage.backends.pathslicing import PathSlicer


def test_pathslicer():
    slicer = PathSlicer("/", "0:2/2:4/4:6")
    assert len(slicer) == 3
    assert slicer.check_config() is None
    assert (
        slicer.get_path("34973274ccef6ab4dfaaf86599792fa9c3fe4689")
        == "/34/97/32/34973274ccef6ab4dfaaf86599792fa9c3fe4689"
    )
    assert (
        slicer.get_directory("34973274ccef6ab4dfaaf86599792fa9c3fe4689") == "/34/97/32"
    )
    assert slicer.get_slices("34973274ccef6ab4dfaaf86599792fa9c3fe4689") == [
        "34",
        "97",
        "32",
    ]

    slicer = PathSlicer("/", "/0:1/0:5/")  # trailing '/' are ignored
    assert slicer.check_config() is None
    assert len(slicer) == 2
    assert (
        slicer.get_path("34973274ccef6ab4dfaaf86599792fa9c3fe4689")
        == "/3/34973/34973274ccef6ab4dfaaf86599792fa9c3fe4689"
    )
    assert (
        slicer.get_directory("34973274ccef6ab4dfaaf86599792fa9c3fe4689") == "/3/34973"
    )
    assert slicer.get_slices("34973274ccef6ab4dfaaf86599792fa9c3fe4689") == [
        "3",
        "34973",
    ]

    # funny one, with steps
    slicer = PathSlicer("/", "0:6:2/1:7:2")
    assert slicer.check_config() is None
    assert slicer.get_slices("123456789".ljust(40, "0")) == ["135", "246"]

    # reverse works too!
    slicer = PathSlicer("/", "-1::-1")
    assert slicer.check_config() is None
    assert slicer.get_slices("34973274ccef6ab4dfaaf86599792fa9c3fe4689") == [
        "34973274ccef6ab4dfaaf86599792fa9c3fe4689"[::-1]
    ]


def test_pathslicer_noop():
    "test the 'empty' pathslicer"
    slicer = PathSlicer("/", "")
    assert len(slicer) == 0
    assert slicer.check_config() is None
    assert (
        slicer.get_path("34973274ccef6ab4dfaaf86599792fa9c3fe4689")
        == "/34973274ccef6ab4dfaaf86599792fa9c3fe4689"
    )


def test_pathslicer_bad_hash():
    slicer = PathSlicer("/", "0:2/2:4/4:6")
    for hexhash in ("0" * 39, "0" * 41, ""):
        with pytest.raises(AssertionError):
            slicer.get_path(hexhash)


def test_pathslicer_check_config():
    with pytest.raises(ValueError):
        PathSlicer("/", "toto")

    with pytest.raises(ValueError):
        PathSlicer("/", "/1:2/a:b/")

    assert PathSlicer("/", "0:40").check_config() is None
    with pytest.raises(ValueError):
        PathSlicer("/", "0:41").check_config()
    assert PathSlicer("/", "40:").check_config() is None
    with pytest.raises(ValueError):
        PathSlicer("/", "41:").check_config()



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from .objstorage_testing import ObjStorageTestFixture


class TestInMemoryObjStorage(ObjStorageTestFixture):
    @pytest.fixture
    def swh_objstorage_config(self):
        return {"cls": "memory"}



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import contextlib
import functools
import logging
import os
from typing import Iterator, Optional

from flask import request
import msgpack

from swh.core.api import RPCServerApp
from swh.core.api import encode_data_server as encode_data
from swh.core.api import error_handler
from swh.core.config import read as config_read
from swh.core.statsd import statsd
from swh.objstorage.exc import Error, ObjCorruptedError, ObjNotFoundError
from swh.objstorage.factory import get_objstorage as get_swhobjstorage
from swh.objstorage.interface import ObjStorageInterface


def timed(f):
    @functools.wraps(f)
    def w(*a, **kw):
        with statsd.timed(
            "swh_objstorage_request_duration_seconds", tags={"endpoint": f.__name__}
        ):
            return f(*a, **kw)

    return w


@contextlib.contextmanager
def timed_context(f_name):
    with statsd.timed(
        "swh_objstorage_request_duration_seconds", tags={"endpoint": f_name}
    ):
        yield


def get_objstorage():
    global objstorage
    if objstorage is None:
        objstorage = get_swhobjstorage(**app.config["objstorage"])

    return objstorage


class ObjStorageServerApp(RPCServerApp):
    client_exception_classes = (ObjNotFoundError, ObjCorruptedError, Error)
    method_decorators = [timed]

    def pre_add(self, kw):
        """Called before the 'add' method."""
        statsd.increment(
            "swh_objstorage_in_bytes_total",
            len(kw["content"]),
            tags={"endpoint": "add_bytes"},
        )

    def post_get(self, ret, kw):
        """Called after the 'get' method."""
        statsd.increment(
            "swh_objstorage_out_bytes_total", len(ret), tags={"endpoint": "get_bytes"}
        )


app = ObjStorageServerApp(
    __name__,
    backend_class=ObjStorageInterface,
    backend_factory=get_objstorage,
)
objstorage = None


@app.errorhandler(Error)
def argument_error_handler(exception):
    return error_handler(exception, encode_data, status_code=400)


@app.errorhandler(Exception)
def my_error_handler(exception):
    return error_handler(exception, encode_data)


@app.route("/")
@timed
def index():
    return "SWH Objstorage API server"


@app.route("/content")
def list_content():
    last_obj_id = request.args.get("last_obj_id")
    if last_obj_id:
        last_obj_id = bytes.fromhex(last_obj_id)
    limit: Optional[str] = request.args.get("limit")
    if limit:
        limit = int(limit)

    def generate() -> Iterator[bytes]:
        with timed_context("list_content"):
            packer = msgpack.Packer(use_bin_type=True)
            for obj in get_objstorage().list_content(last_obj_id, limit=limit):
                yield packer.pack(obj)

    return app.response_class(generate())


api_cfg = None


def load_and_check_config(config_file):
    """Check the minimal configuration is set to run the api or raise an
       error explanation.

    Args:
        config_file (str): Path to the configuration file to load

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if not config_file:
        raise EnvironmentError("Configuration file must be defined")

    if not os.path.exists(config_file):
        raise FileNotFoundError("Configuration file %s does not exist" % (config_file,))

    cfg = config_read(config_file)
    return validate_config(cfg)


def validate_config(cfg):
    """Check the minimal configuration is set to run the api or raise an
       explanatory error.

    Args:
        cfg (dict): Loaded configuration.

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if "objstorage" not in cfg:
        raise KeyError("Invalid configuration; missing objstorage config entry")

    missing_keys = []
    vcfg = cfg["objstorage"]
    if "cls" not in vcfg:
        raise KeyError("Invalid configuration; missing cls config entry")

    cls = vcfg["cls"]
    if cls == "pathslicing":
        for key in ("root", "slicing"):
            v = vcfg.get(key)
            if v is None:
                missing_keys.append(key)

        if missing_keys:
            raise KeyError(
                "Invalid configuration; missing %s config entry"
                % (", ".join(missing_keys),)
            )

    return cfg


def make_app_from_configfile():
    """Load configuration and then build application to run"""
    global api_cfg
    if not api_cfg:
        config_path = os.environ.get("SWH_CONFIG_FILENAME")
        api_cfg = load_and_check_config(config_path)
        app.config.update(api_cfg)
    handler = logging.StreamHandler()
    app.logger.addHandler(handler)
    return app


if __name__ == "__main__":
    print("Deprecated. Use swh-objstorage")



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import os
from urllib.parse import urlparse

import pytest
from requests.utils import get_encoding_from_headers
import requests_mock
from requests_mock.contrib import fixture

from swh.objstorage.backends.pathslicing import PathSlicer
from swh.objstorage.exc import ObjCorruptedError
from swh.objstorage.objstorage import compute_hash, decompressors
from swh.objstorage.tests.objstorage_testing import ObjStorageTestFixture


class PathDict:
    """A dict-like object that handles "path-like" keys in a recursive dict
    structure.

    For example:

        >>> a = PathDict()
        >>> a['path/to/file'] = 'some file content'

    will create a dict structure (in self.data) like:

        >>> print(a.data)
        {'path': {'to': {'file': 'some file content'}}}
        >>> 'path/to/file' in a
        True

    This is a helper class for the FilerRequestsMock below.
    """

    def __init__(self):
        self.data = {}

    def __setitem__(self, key, value):
        if key.endswith("/"):
            raise ValueError("Nope")
        if key.startswith("/"):
            key = key[1:]
        path = key.split("/")
        resu = self.data
        for p in path[:-1]:
            resu = resu.setdefault(p, {})
        resu[path[-1]] = value

    def __getitem__(self, key):
        assert isinstance(key, str)
        if key == "/":
            return self.data

        if key.startswith("/"):
            key = key[1:]
        if key.endswith("/"):
            key = key[:-1]

        path = key.split("/")
        resu = self.data
        for p in path:
            resu = resu[p]
        return resu

    def __delitem__(self, key):
        if key.startswith("/"):
            key = key[1:]
        if key.endswith("/"):
            key = key[:-1]
        path = key.split("/")
        resu = self.data
        for p in path[:-1]:
            resu = resu.setdefault(p, {})
        del resu[path[-1]]

    def __contains__(self, key):
        if key == "/":
            # always consider we have the 'root' directory
            return True
        try:
            self[key]
            return True
        except KeyError:
            return False


class FilerRequestsMock:
    """This is a requests_mock based mock for the seaweedfs Filer API

    It does not implement the whole API, only the parts required to make the
    HttpFiler (used by SeaweedFilerObjStorage) work.

    It stores the files in a dict-based structure, eg. the file
    '0a/32/0a3245983255' will be stored in a dict like:

        {'0a': {'32': {'0a3245983255': b'content'}}}

    It uses the PathDict helper class to make it a bit easier to handle this
    dict structure.

    """

    MODE_DIR = 0o20000000771
    MODE_FILE = 0o660

    def __init__(self, url):
        self.url = url
        self.root_path = urlparse(url).path
        self.content = PathDict()
        self.requests_mock = fixture.Fixture()
        self.requests_mock.setUp()
        self.requests_mock.register_uri(
            requests_mock.GET, requests_mock.ANY, content=self.get_cb
        )
        self.requests_mock.register_uri(
            requests_mock.POST, requests_mock.ANY, content=self.post_cb
        )
        self.requests_mock.register_uri(
            requests_mock.HEAD, requests_mock.ANY, content=self.head_cb
        )
        self.requests_mock.register_uri(
            requests_mock.DELETE, requests_mock.ANY, content=self.delete_cb
        )

    def relpath(self, path):
        if path.startswith(self.root_path):
            return os.path.relpath(path, self.root_path)

    def head_cb(self, request, context):
        if request.path not in self.content:
            context.status_code = 404
            return b"Not Found"

    def get_cb(self, request, context):
        if self.head_cb(request, context) == b"Not Found":
            return

        if "limit" in request.qs:
            limit = int(request.qs["limit"][0])
            assert limit > 0
        else:
            limit = None

        if "lastfilename" in request.qs:
            lastfilename = request.qs["lastfilename"][0]
        else:
            lastfilename = None

        content = self.content[request.path]
        if isinstance(content, dict):
            # it's a directory; list its content
            entries = sorted(content.keys())
            if not entries:
                text = json.dumps(
                    {
                        "EmptyFolder": True,
                        "Entries": None,
                        "LastFileName": lastfilename,
                        "Limit": limit,
                        "Path": request.path,
                        "ShouldDisplayLoadMore": False,
                    }
                )
                encoding = get_encoding_from_headers(request.headers) or "utf-8"
                return text.encode(encoding)

            if lastfilename:
                # exclude all filenames up to lastfilename
                entries = [k for k in entries if k > lastfilename]

            thereismore = False
            if limit and len(entries) > limit:
                entries = entries[:limit]
                thereismore = True

            resp_entries = []
            for entry in entries:
                fullpath = os.path.join(request.path, entry)
                if isinstance(self.content[fullpath], dict):
                    mode = self.MODE_DIR
                else:
                    mode = self.MODE_FILE
                resp_entries.append(
                    {
                        "FullPath": fullpath,
                        "Mode": mode,
                    }
                )

            if resp_entries:
                lastfilename = resp_entries[-1]["FullPath"].split("/")[-1]
            else:
                lastfilename = None
            text = json.dumps(
                {
                    "Path": request.path,
                    "Limit": limit,
                    "LastFileName": lastfilename,
                    "ShouldDisplayLoadMore": thereismore,
                    "Entries": resp_entries,
                }
            )
            encoding = get_encoding_from_headers(request.headers) or "utf-8"
            return text.encode(encoding)
        else:
            # return the actual file content
            return content

    def post_cb(self, request, context):
        from requests_toolbelt.multipart import decoder

        multipart_data = decoder.MultipartDecoder(
            request.body, request.headers["content-type"]
        )
        part = multipart_data.parts[0]
        self.content[request.path] = part.content

    def delete_cb(self, request, context):
        del self.content[request.path]


class TestSeaweedObjStorage(ObjStorageTestFixture):
    compression = "none"
    url = "http://127.0.0.1/test/"
    slicing = ""

    @pytest.fixture
    def swh_objstorage_config(self):
        self.mock = FilerRequestsMock(url=self.url)
        return {
            "cls": "seaweedfs",
            "url": self.url,
            "compression": self.compression,
            "slicing": self.slicing,
        }

    def fill_objstorage(self, num_objects):
        # override default implelentation to speed things up a bit, shortcuting
        # the HTTP request path to put objects directly in the objstorage
        # mocker.

        path = self.storage._path
        all_ids = []
        for i in range(num_objects):
            content = b"content %d" % i
            obj_id = compute_hash(content)
            self.mock.content[path(obj_id)] = self.storage.compress(content)
            all_ids.append({"sha1": obj_id})
        all_ids.sort(key=lambda d: d["sha1"])
        return all_ids

    def test_compression(self):
        content, obj_id = self.hash_content(b"test compression")
        self.storage.add(content, obj_id=obj_id)

        raw_content = self.storage.wf.get(self.storage._path(obj_id))
        if self.compression == "none":
            assert raw_content == content
        else:
            assert raw_content != content
        d = decompressors[self.compression]()
        assert d.decompress(raw_content) == content
        assert d.unused_data == b""

    def test_trailing_data_on_stored_blob(self):
        content, obj_id = self.hash_content(b"test content without garbage")
        self.storage.add(content, obj_id=obj_id)

        path = self.storage._path(obj_id)
        self.mock.content[path] += b"trailing garbage"

        with pytest.raises(ObjCorruptedError) as e:
            self.storage.check(obj_id)

        if self.compression != "none":
            assert "trailing data found" in e.value.args[0]

    def test_slicing(self):
        slicer = PathSlicer(urlparse(self.url).path, self.slicing)
        for i in range(20):
            content, obj_id = self.hash_content(b"test slicing %i" % i)
            self.storage.add(content, obj_id=obj_id)
            assert slicer.get_path(obj_id.hex()) in self.mock.content


class TestSeaweedObjStorageWithCompression(TestSeaweedObjStorage):
    compression = "lzma"


class TestSeaweedObjStorageWithSlicing1(TestSeaweedObjStorage):
    num_objects = 12000
    slicing = "0:2"


class TestSeaweedObjStorageWithSlicing2(TestSeaweedObjStorage):
    num_objects = 12000
    slicing = "0:1/1:2"


class TestSeaweedObjStorageWithSmallBatch(TestSeaweedObjStorage):
    num_objects = 120

    @pytest.fixture(autouse=True)
    def batchsize(self, swh_objstorage):
        swh_objstorage.wf.batchsize = 1


class TestSeaweedObjStorageWithSlicing1AndSmallBatch(
    TestSeaweedObjStorageWithSmallBatch
):
    slicing = "0:2"


class TestSeaweedObjStorageWithSlicing2AndSmallBatch(
    TestSeaweedObjStorageWithSmallBatch
):
    slicing = "0:1/1:2"


class TestSeaweedObjStorageWithNoPath(TestSeaweedObjStorage):
    url = "http://127.0.0.1/"



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.objstorage.factory import get_objstorage


@pytest.fixture
def swh_objstorage_config():
    """Fixture that returns a dictionary containing the configuration
    required to instantiate an object storage.

    Unless the fixture gets overridden, the configuration for creating an
    object storage in memory is returned.

    See :func:`swh.objstorage.factory.get_objstorage` for more details.
    """
    return {"cls": "memory"}


@pytest.fixture
def swh_objstorage(swh_objstorage_config):
    """Fixture that instantiates an object storage based on the configuration
    returned by the ``swh_objstorage_config`` fixture.
    """
    return get_objstorage(**swh_objstorage_config)



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import timedelta
import logging
from typing import Iterator, Optional
from urllib.parse import urljoin

import requests

from swh.model import hashutil
from swh.objstorage.constants import ID_HASH_ALGO
from swh.objstorage.exc import (
    NonIterableObjStorage,
    ObjNotFoundError,
    ReadOnlyObjStorage,
)
from swh.objstorage.interface import CompositeObjId, ObjId
from swh.objstorage.objstorage import (
    DEFAULT_LIMIT,
    CompressionFormat,
    ObjStorage,
    objid_to_default_hex,
)

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.ERROR)


class HTTPReadOnlyObjStorage(ObjStorage):
    """Simple ObjStorage retrieving objects from an HTTP server.

    For example, can be used to retrieve objects from S3:

    objstorage:
      cls: http
      url: https://softwareheritage.s3.amazonaws.com/content/
    """

    def __init__(self, url=None, compression: CompressionFormat = "none", **kwargs):
        super().__init__(**kwargs)
        self.session = requests.sessions.Session()
        self.root_path = url
        if not self.root_path.endswith("/"):
            self.root_path += "/"
        self.compression = compression

    def check_config(self, *, check_write):
        """Check the configuration for this object storage"""
        return True

    def __contains__(self, obj_id: ObjId) -> bool:
        resp = self.session.head(self._path(obj_id))
        return resp.status_code == 200

    def __iter__(self) -> Iterator[CompositeObjId]:
        raise NonIterableObjStorage("__iter__")

    def __len__(self):
        raise NonIterableObjStorage("__len__")

    def add(self, content: bytes, obj_id: ObjId, check_presence: bool = True) -> None:
        raise ReadOnlyObjStorage("add")

    def delete(self, obj_id: ObjId):
        raise ReadOnlyObjStorage("delete")

    def restore(self, content: bytes, obj_id: ObjId) -> None:
        raise ReadOnlyObjStorage("restore")

    def list_content(
        self,
        last_obj_id: Optional[ObjId] = None,
        limit: Optional[int] = DEFAULT_LIMIT,
    ) -> Iterator[CompositeObjId]:
        raise NonIterableObjStorage("__len__")

    def get(self, obj_id: ObjId) -> bytes:
        try:
            resp = self.session.get(self._path(obj_id))
            resp.raise_for_status()
        except Exception:
            raise ObjNotFoundError(obj_id)

        return self.decompress(resp.content, objid_to_default_hex(obj_id))

    def download_url(
        self,
        obj_id: ObjId,
        content_disposition: Optional[str] = None,
        expiry: Optional[timedelta] = None,
    ) -> Optional[str]:
        return self._path(obj_id)

    def _hash(self, obj_id: ObjId) -> bytes:
        if isinstance(obj_id, dict):
            return obj_id[ID_HASH_ALGO]
        else:
            return obj_id

    def _path(self, obj_id):
        return urljoin(self.root_path, hashutil.hash_to_hex(self._hash(obj_id)))



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from contextlib import contextmanager, nullcontext
from functools import partial
import logging
from threading import Event, Thread
import time
from typing import Callable, ContextManager, Iterator, Optional, Protocol, Tuple

import psycopg

from .database import Database

logger = logging.getLogger(__name__)


class IdleHandler(Thread):
    """Call the `callback` after being idle for `timeout` seconds."""

    def __init__(self, name: str, timeout: float, callback: Callable[[], None]):
        super().__init__(name=f"IdleHandler-{name}")
        self.timeout = timeout
        self.callback = callback
        self.deadline = time.monotonic() + timeout
        self.quiesced = Event()
        self.terminated = Event()

    def quiesce(self):
        """Quiesce the timeout.

        This should generally be used via the :func:`quiesce_then_reset` context
        manager, which wraps a block of code to quiesce the timeout while the
        code runs, then resets the timeout on completion.

        """
        self.quiesced.set()

    def reset(self):
        """Reset the timeout clock.

        This should generally be used via the :func:`quiesce_then_reset` context
        manager, which wraps a block of code to quiesce the timeout while the
        code runs, then resets the timeout on completion.
        """
        self.deadline = time.monotonic() + self.timeout
        self.quiesced.clear()

    @contextmanager
    def quiesce_then_reset(self):
        """Wrap a block of code to quiesce the timeout while the code runs,
        then reset the timeout on completion.
        """
        self.quiesce()
        yield
        self.reset()

    def join(self, timeout=None):
        """Gracefully terminate the thread."""
        self.terminated.set()
        # Trigger exit from the main loop by setting the quiesced event
        self.quiesced.set()
        return super().join(timeout)

    def run(self):
        while True:
            # Wait at least 1 second when paused
            wait_for = max(self.deadline - time.monotonic(), 1)
            quiesced = self.quiesced.wait(timeout=wait_for)
            if self.terminated.is_set():
                break
            if quiesced:
                time.sleep(0.1)
                continue
            if time.monotonic() > self.deadline:
                break

        if not self.terminated.is_set():
            logger.debug("Idle timeout reached, calling idle callback")
            self.callback()


class ShardIdleTimeoutCallback(Protocol):
    """A function which takes a :class:`RWShard` as `shard` argument, used as
    idle timeout callback for :class:`RWShard`."""

    def __call__(self, shard: "RWShard") -> None:
        ...


class RWShard(Database):
    def __init__(
        self,
        name: str,
        base_dsn: str,
        shard_max_size: int,
        application_name: Optional[str] = None,
        idle_timeout_cb: Optional[ShardIdleTimeoutCallback] = None,
        idle_timeout: Optional[float] = 5,
        **kwargs,
    ):
        self._name = name
        if application_name is None:
            application_name = f"SWH Winery RW Shard {name}"
        super().__init__(dsn=base_dsn, application_name=application_name)
        self.create()
        self.size = self.total_size()
        self.limit = shard_max_size

        self.idle_handler: Optional[IdleHandler] = None
        self.quiesce_then_reset_idle: Callable[[], ContextManager] = nullcontext

        if idle_timeout and idle_timeout_cb:
            self.idle_handler = IdleHandler(
                name=name,
                timeout=idle_timeout,
                callback=partial(idle_timeout_cb, shard=self),
            )
            self.idle_handler.start()
            self.quiesce_then_reset_idle = self.idle_handler.quiesce_then_reset

    def disable_idle_handler(self):
        if thread := getattr(self, "idle_handler"):
            thread.join()
            self.idle_handler = None
            self.quiesce_then_reset_idle = nullcontext

    @property
    def name(self) -> str:
        return self._name

    @property
    def table_name(self) -> str:
        return f"shard_{self._name}"

    def is_full(self) -> bool:
        return self.size >= self.limit

    def create(self) -> None:
        with self.pool.connection() as db:
            db.execute(
                f"CREATE TABLE IF NOT EXISTS {self.table_name} "
                "(LIKE shard_template INCLUDING ALL) "
                "WITH (autovacuum_enabled = false)"
            )

    def drop(self) -> None:
        with self.pool.connection() as db:
            db.execute(f"DROP TABLE {self.table_name}")

    def total_size(self) -> int:
        with self.pool.connection() as db, db.cursor() as c:
            c.execute(f"SELECT SUM(LENGTH(content)) FROM {self.table_name}")
            size = c.fetchone()[0]
            if size is None:
                return 0
            else:
                return size

    def add(self, db: psycopg.Connection, obj_id: bytes, content: bytes) -> None:
        with self.quiesce_then_reset_idle():
            cur = db.execute(
                f"INSERT INTO {self.table_name} (key, content) "
                "VALUES (%s, %s) "
                "ON CONFLICT (key) DO NOTHING",
                (obj_id, content),
                binary=True,
            )
            if cur.rowcount:
                self.size += len(content)

    def get(self, obj_id: bytes) -> Optional[bytes]:
        with self.pool.connection() as db, db.cursor() as c:
            c.execute(
                f"SELECT content FROM {self.table_name} WHERE key = %s",
                (obj_id,),
                binary=True,
            )
            if c.rowcount == 0:
                return None
            else:
                return c.fetchone()[0]

    def delete(self, obj_id: bytes) -> None:
        with self.pool.connection() as db, db.cursor() as c:
            c.execute(f"DELETE FROM {self.table_name} WHERE key = %s", (obj_id,))
            if c.rowcount == 0:
                raise KeyError(obj_id)

    def all(self) -> Iterator[Tuple[bytes, bytes]]:
        with self.pool.connection() as db, db.cursor() as c:
            with c.copy(
                f"COPY {self.table_name} (key, content) TO STDOUT (FORMAT BINARY)"
            ) as copy:
                copy.set_types(["bytea", "bytea"])
                yield from copy.rows()

    def count(self) -> int:
        with self.pool.connection() as db, db.cursor() as c:
            c.execute(f"SELECT COUNT(*) FROM {self.table_name}")
            return c.fetchone()[0]



import sys

import pytest

pytest_plugins = ["swh.objstorage.pytest_plugin"]


def pytest_configure(config):
    config.addinivalue_line("markers", "shard_max_size: winery backend")
    config.addinivalue_line(
        "markers", "pack_immediately(bool): whether winery should pack immediately"
    )
    config.addinivalue_line(
        "markers",
        "clean_immediately(bool): whether the winery packer should clean rw "
        "shards immediately",
    )
    config.addinivalue_line(
        "markers", "use_benchmark_flags: use the --winery-bench-* CLI flags"
    )
    config.addinivalue_line(
        "markers",
        (
            "all_compression_methods: "
            "test all compression methods instead of only the most common ones"
        ),
    )
    config.addinivalue_line(
        "markers", "skip_on_cloud: skip test on cloud implementations"
    )


def pytest_addoption(parser):
    if sys.version_info >= (3, 9):
        import argparse

        action = argparse.BooleanOptionalAction
        default = True
    else:
        action = "store_true"
        default = False

    parser.addoption(
        "--winery-bench-pack-immediately",
        action=action,
        help="Pack objects synchronously in benchmark",
        default=default,
    )

    parser.addoption(
        "--winery-bench-remove-pool",
        action=action,
        help="Remove Ceph pool before and after tests",
        default=default,
    )

    parser.addoption(
        "--winery-bench-remove-images",
        action=action,
        help="Remove Ceph images after tests",
        default=default,
    )

    parser.addoption(
        "--winery-bench-rbd-pool",
        help="RBD pool for benchmark",
        default="winery-benchmark-shards",
    )

    parser.addoption(
        "--winery-bench-output-directory",
        help="Directory in which the performance results are stored",
        default="/tmp/winery",
    )
    parser.addoption(
        "--winery-bench-rw-workers",
        type=int,
        help="Number of Read/Write workers",
        default=1,
    )
    parser.addoption(
        "--winery-bench-ro-workers",
        type=int,
        help="Number of Readonly workers",
        default=1,
    )
    parser.addoption(
        "--winery-bench-pack-workers",
        type=int,
        help="Number of Pack workers",
        default=1,
    )
    parser.addoption(
        "--winery-bench-duration",
        type=int,
        help="Duration of the benchmarks in seconds",
        default=1,
    )
    parser.addoption(
        "--winery-bench-shard-max-size",
        type=int,
        help="Size of the shard in bytes",
        default=10 * 1024 * 1024,
    )
    parser.addoption(
        "--winery-bench-stats-interval",
        type=int,
        help="Interval between stat computations (seconds)",
        default=5 * 60,
    )
    parser.addoption(
        "--winery-bench-ro-worker-max-request",
        type=int,
        help="Number of requests a ro worker performs",
        default=1,
    )
    parser.addoption(
        "--winery-bench-throttle-read",
        type=int,
        help="Maximum number of bytes per second read",
        default=100 * 1024 * 1024,
    )
    parser.addoption(
        "--winery-bench-throttle-write",
        type=int,
        help="Maximum number of bytes per second write",
        default=100 * 1024 * 1024,
    )
    parser.addoption(
        "--all-compression-methods",
        action="store_true",
        default=False,
        help="Test all compression methods",
    )


def pytest_runtest_setup(item):
    if item.get_closest_marker("all_compression_methods"):
        if not item.config.getoption("--all-compression-methods"):
            pytest.skip("`--all-compression-methods` has not been specified")
    if item.get_closest_marker("skip_on_cloud"):
        if item.parent and "Cloud" in item.parent.name:
            pytest.skip("skipping on cloud implementation")



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import Counter
from dataclasses import asdict, dataclass
import datetime
from functools import partial
import logging
import os
import shutil
import threading
import time
from typing import Any, Dict

from click.testing import CliRunner
import pytest
from pytest_postgresql import factories
import yaml

from swh.core.db.db_utils import initialize_database_for_module
import swh.objstorage.backends.winery.objstorage
from swh.objstorage.backends.winery.objstorage import (
    WineryObjStorage,
    cleanup_rw_shard,
    deleted_objects_cleaner,
    pack,
    rw_shard_cleaner,
    shard_packer,
    stop_after_shards,
)
from swh.objstorage.backends.winery.sharedbase import ShardState, SharedBase
from swh.objstorage.backends.winery.sleep import sleep_exponential
from swh.objstorage.backends.winery.stats import Stats
from swh.objstorage.backends.winery.throttler import (
    BandwidthCalculator,
    IOThrottler,
    LeakyBucket,
    Throttler,
)
from swh.objstorage.cli import swh_cli_group
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.factory import get_objstorage
from swh.objstorage.objstorage import compute_hash
from swh.objstorage.utils import call_async

from .winery_benchmark import (
    Bench,
    PackWorker,
    RBDWorker,
    ROWorker,
    RWShardCleanerWorker,
    RWWorker,
    StatsPrinter,
    WorkerKind,
    work,
)
from .winery_testing_helpers import FileBackedPool, PoolHelper

logger = logging.getLogger(__name__)


@pytest.fixture
def needs_ceph():
    ceph = shutil.which("ceph")

    if not ceph or os.environ.get("USE_CEPH", "yes") != "yes":
        pytest.skip("the ceph CLI was not found")


@pytest.fixture
def cli_runner(capsys):
    "Run click commands with log capture disabled"

    class CapsysDisabledCliRunner(CliRunner):
        def invoke(self, *args, **kwargs):
            with capsys.disabled():
                return super().invoke(*args, **kwargs)

    return CapsysDisabledCliRunner()


@pytest.fixture
def remove_pool(request, pytestconfig):
    if os.environ.get("CEPH_HARDCODE_POOL"):
        return False
    marker = request.node.get_closest_marker("use_benchmark_flags")
    if marker is None:
        return True

    return pytestconfig.getoption("--winery-bench-remove-pool")


@pytest.fixture
def remove_images(request, pytestconfig):
    if os.environ.get("CEPH_HARDCODE_POOL"):
        return False
    marker = request.node.get_closest_marker("use_benchmark_flags")
    if marker is None:
        return True

    return pytestconfig.getoption("--winery-bench-remove-images")


@pytest.fixture
def rbd_pool_name(request, pytestconfig):
    if os.environ.get("CEPH_HARDCODE_POOL"):
        return os.environ["CEPH_HARDCODE_POOL"]
    marker = request.node.get_closest_marker("use_benchmark_flags")
    if marker is None:
        return "winery-test-shards"

    return pytestconfig.getoption("--winery-bench-rbd-pool")


@pytest.fixture
def rbd_map_options():
    return os.environ.get("RBD_MAP_OPTIONS", "")


@pytest.fixture
def ceph_pool(remove_pool, remove_images, rbd_pool_name, rbd_map_options, needs_ceph):
    pool = PoolHelper(
        shard_max_size=10 * 1024 * 1024,
        rbd_pool_name=rbd_pool_name,
        rbd_map_options=rbd_map_options,
    )
    if remove_pool:
        pool.remove()
        pool.pool_create()
    else:
        logger.info("Not removing pool")

    yield pool

    if remove_images or remove_pool:
        pool.images_remove()
    else:
        logger.info("Not removing images")

    if remove_pool:
        pool.remove()
    else:
        logger.info("Not removing pool")


@pytest.fixture
def file_backed_pool(mocker, tmp_path, shard_max_size, rbd_pool_name):
    FileBackedPool.set_base_directory(tmp_path)
    mocker.patch(
        "swh.objstorage.backends.winery.roshard.Pool",
        new=FileBackedPool,
    )
    mocker.patch(
        "swh.objstorage.tests.winery_benchmark.Pool",
        new=FileBackedPool,
    )
    pool = FileBackedPool(shard_max_size=10 * 1024 * 1024, rbd_pool_name=rbd_pool_name)
    pool.image_unmap_all()
    yield pool


def pytest_generate_tests(metafunc):
    if "image_pool" in metafunc.fixturenames:
        metafunc.parametrize(
            "image_pool", ["ceph_pool", "file_backed_pool"], indirect=True
        )


@pytest.fixture
def image_pool(request):
    return request.getfixturevalue(request.param)


winery_postgresql_proc = factories.postgresql_proc(
    load=[
        partial(
            initialize_database_for_module,
            modname="objstorage.backends.winery",
            version=SharedBase.current_version,
        ),
    ],
)

winery_postgresql = factories.postgresql("winery_postgresql_proc")


@pytest.fixture
def postgresql_dsn(winery_postgresql):
    return winery_postgresql.info.dsn


@pytest.fixture
def shard_max_size(request) -> int:
    marker = request.node.get_closest_marker("shard_max_size")
    if marker is None:
        return 1024
    else:
        return marker.args[0]


@pytest.fixture
def pack_immediately(request) -> bool:
    marker = request.node.get_closest_marker("pack_immediately")
    if marker is None:
        return True
    else:
        return marker.args[0]


@pytest.fixture
def clean_immediately(request) -> bool:
    marker = request.node.get_closest_marker("clean_immediately")
    if marker is None:
        return True
    else:
        return marker.args[0]


@pytest.fixture
def storage(
    shard_max_size,
    pack_immediately,
    clean_immediately,
    rbd_pool_name,
    rbd_map_options,
    postgresql_dsn,
):
    storage = get_objstorage(
        cls="winery",
        base_dsn=postgresql_dsn,
        shard_max_size=shard_max_size,
        throttle_write=200 * 1024 * 1024,
        throttle_read=100 * 1024 * 1024,
        pack_immediately=pack_immediately,
        clean_immediately=clean_immediately,
        rbd_pool_name=rbd_pool_name,
        rbd_map_options=rbd_map_options,
    )
    assert isinstance(storage, WineryObjStorage)
    logger.debug("Instantiated storage %s on rbd pool %s", storage, rbd_pool_name)
    yield storage
    storage.on_shutdown()
    names = [
        thread.name
        for thread in threading.enumerate()
        if thread.name.startswith("IdleHandler")
    ]
    assert not names, f"Some IdleHandlers are still alive: {','.join(names)}"


@pytest.fixture
def winery(storage):
    return storage.winery


def test_winery_sharedbase(winery):
    base = winery.base
    shard1 = winery.shard.name
    assert shard1 is not None
    assert shard1 == base.locked_shard

    id1 = base.locked_shard_id
    assert id1 is not None
    assert id1 == base.locked_shard_id

    assert base.get_shard_state(shard1) == ShardState.WRITING

    winery.release_shard()

    assert winery.base._locked_shard is None
    assert base.get_shard_state(shard1) == ShardState.STANDBY

    shard2 = winery.base.locked_shard

    assert shard1 == shard2, "Locked a different shard?"
    assert base.get_shard_state(shard1) == ShardState.WRITING


def test_winery_add_get(winery):
    shard = winery.base.locked_shard
    content = b"SOMETHING"
    obj_id = compute_hash(content, "sha256")
    assert (
        obj_id.hex()
        == "866878b165607851782d8d233edf0c261172ff67926330d3bbd10c705b92d24f"
    )
    winery.add(content=content, obj_id=obj_id)
    winery.add(content=content, obj_id=obj_id)
    winery.add(content=content, obj_id=obj_id, check_presence=False)
    assert winery.base.locked_shard == shard
    assert winery.get(obj_id) == content
    with pytest.raises(ObjNotFoundError):
        winery.get(b"unknown")
    winery.shard.drop()


def test_winery_add_concurrent(winery, mocker):
    num_threads = 4

    class ManualReleaseSharedBase(SharedBase):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.release_obj_id = threading.Event()

        def record_new_obj_id(self, *args, **kwargs):
            ret = super().record_new_obj_id(*args, **kwargs)
            self.release_obj_id.wait()
            return ret

    mocker.patch(
        "swh.objstorage.backends.winery.objstorage.SharedBase", ManualReleaseSharedBase
    )

    content = b"test_concurrency"
    obj_id = compute_hash(content, "sha256")

    def add_object(my_storage):
        my_storage.add(content=content, obj_id=obj_id)

        assert my_storage.get(obj_id) == content

    storages = [get_objstorage(cls="winery", **winery.args) for _ in range(num_threads)]

    threads = [
        threading.Thread(target=add_object, args=[storage]) for storage in storages
    ]
    for thread in threads:
        thread.start()

    for storage in reversed(storages):
        storage.winery.base.release_obj_id.set()

    for thread in threads:
        thread.join()

    assert winery.get(obj_id) == content
    assert sum(1 for _ in winery.base.list_shards()) >= num_threads

    for storage in storages:
        assert isinstance(storage, WineryObjStorage)
        storage.on_shutdown()


@pytest.mark.shard_max_size(1)
def test_winery_add_and_pack(winery, mocker):
    mocker.patch("swh.objstorage.backends.winery.objstorage.pack", return_value=True)
    shard = winery.base.locked_shard
    content = b"SOMETHING"
    winery.add(content=content, obj_id=compute_hash(content, "sha256"))
    assert winery.base.locked_shard != shard
    assert len(winery.packers) == 1
    packer = winery.packers[0]
    packer.join()
    assert packer.exitcode == 0


def test_winery_delete_on_rwshard(winery):
    shard = winery.base.locked_shard
    content = b"SOMETHING"
    obj_id = compute_hash(content, "sha256")
    winery.add(content=content, obj_id=obj_id)
    assert winery.base.locked_shard == shard
    assert winery.get(obj_id) == content
    winery.delete(obj_id)
    with pytest.raises(ObjNotFoundError):
        winery.get(obj_id)


@pytest.mark.shard_max_size(1)
@pytest.mark.pack_immediately(True)
def test_winery_delete_on_roshard(winery, file_backed_pool):
    shard = winery.base.locked_shard
    content = b"SOMETHING"
    obj_id = compute_hash(content, "sha256")
    winery.add(content=content, obj_id=obj_id)
    assert winery.base.locked_shard != shard
    assert winery.packers
    for packer in winery.packers:
        packer.join()
    assert winery.get(obj_id) == content
    # This will only mark as deleted in SharedBase
    winery.delete(obj_id)
    assert len(list(winery.base.deleted_objects())) == 1
    # We still should not be able to access it
    with pytest.raises(ObjNotFoundError):
        winery.get(obj_id)
    # The content is still present in the roshard image at this point
    image_path = file_backed_pool.image_path(shard)
    with open(image_path, "rb") as image:
        assert b"SOMETHING" in image.read()
    # Perform cleanup
    file_backed_pool.image_unmap(shard)
    file_backed_pool.image_map(shard, "rw")
    deleted_objects_cleaner(winery.base, file_backed_pool, stop_running=lambda: False)
    assert len(list(winery.base.deleted_objects())) == 0
    with open(image_path, "rb") as image:
        assert b"SOMETHING" not in image.read()


@pytest.mark.shard_max_size(20)
@pytest.mark.pack_immediately(True)
def test_winery_deleted_objects_cleaner_handles_exception(
    winery, file_backed_pool, mocker
):
    from swh.objstorage.backends.winery import objstorage as winery_objstorage

    from ..backends.winery.roshard import ROShard

    # Add two objects
    shard = winery.base.locked_shard
    content1 = b"PINOT GRIS"
    obj_id1 = compute_hash(content1, "sha256")
    winery.add(content=content1, obj_id=obj_id1)
    content2 = b"CHARDONNAY"
    obj_id2 = compute_hash(content2, "sha256")
    winery.add(content=content2, obj_id=obj_id2)

    # This should be enough bytes to trigger packing
    for packer in winery.packers:
        packer.join()

    # We should only have one roshard
    assert len(file_backed_pool.image_list()) == 1

    # This will only mark as deleted in SharedBase for the time being
    winery.delete(obj_id1)
    winery.delete(obj_id2)
    assert len(list(winery.base.deleted_objects())) == 2

    # The content is still present in the roshard image at this point
    image_path = file_backed_pool.image_path(shard)

    # Setup so we get an exception on the second object
    already_called = False
    orig_roshard_delete = ROShard.delete

    def roshard_delete_side_effect(pool, shard_name, obj_id):
        nonlocal already_called
        print(already_called)
        if already_called:
            raise OSError("Unable to write to pool")
        orig_roshard_delete(pool, shard_name, obj_id)
        already_called = True
        return None

    mocker.patch.object(
        winery_objstorage.ROShard, "delete", side_effect=roshard_delete_side_effect
    )

    # Let’s run the cleaner
    file_backed_pool.image_unmap(shard)
    file_backed_pool.image_map(shard, "rw")
    with pytest.raises(OSError):
        winery_objstorage.deleted_objects_cleaner(
            winery.base, file_backed_pool, stop_running=lambda: False
        )

    # We should only have one remaining object to delete
    assert len(list(winery.base.deleted_objects())) == 1

    # We should have only the content of one of the objects still in the roshard
    with open(image_path, "rb") as image:
        image_content = image.read()
        presences = [content1 in image_content, content2 in image_content]
        assert sorted(presences) == [False, True]


def test_winery_get_shard_info(winery):
    assert winery.base.get_shard_info(1234) is None
    assert winery.base.get_shard_state("nothing") is None


def test_winery_base_record_shard_mapped(winery):
    # Lock a shard
    shard_name, shard_id = winery.base.create_shard(new_state=ShardState.PACKED)

    assert {"test"} == winery.base.record_shard_mapped(host="test", name=shard_name)
    assert {"test"} == winery.base.record_shard_mapped(host="test", name=shard_name)
    assert {"test", "test2"} == winery.base.record_shard_mapped(
        host="test2", name=shard_name
    )


@pytest.mark.shard_max_size(10 * 1024 * 1024)
@pytest.mark.clean_immediately(False)
def test_winery_pack(winery, image_pool):
    shard = winery.base.locked_shard
    content = b"SOMETHING"
    obj_id = compute_hash(content, "sha256")
    winery.add(content=content, obj_id=obj_id)
    winery.base.set_shard_state(ShardState.FULL)
    winery.base.shard_packing_starts(shard)

    assert pack(shard, **winery.args)
    assert winery.base.get_shard_state(shard) == ShardState.PACKED

    assert cleanup_rw_shard(shard, **winery.args)
    assert winery.base.get_shard_state(shard) == ShardState.READONLY


@pytest.mark.shard_max_size(1024 * 1024)
@pytest.mark.pack_immediately(True)
def test_winery_writer_pack_immediately_true(image_pool, storage):
    shard = storage.winery.base.locked_shard

    for i in range(1024):
        content = i.to_bytes(1024, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    assert storage.winery.packers
    for packer in storage.winery.packers:
        packer.join()

    assert storage.winery.base.locked_shard != shard

    assert storage.winery.base.get_shard_state(shard) == ShardState.READONLY


@pytest.mark.shard_max_size(1024 * 1024)
@pytest.mark.pack_immediately(False)
def test_winery_writer_pack_immediately_false(storage):
    shard = storage.winery.base.locked_shard

    for i in range(1024):
        content = i.to_bytes(1024, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    assert storage.winery.base.locked_shard != shard
    assert not storage.winery.packers

    assert storage.winery.base.get_shard_state(shard) == ShardState.FULL


@pytest.mark.parametrize(
    "min_duration,factor,max_duration,expected",
    (
        (1, 2, 10, [1, 2, 4, 8, 10, 10]),
        (10, 1.5, 20, [10, 15.0, 20.0, 20.0]),
        (20, 1.3, 10, [10, 10, 10]),
    ),
)
def test_winery_sleep_exponential(mocker, min_duration, factor, max_duration, expected):
    calls = []

    def mocked_sleep(t: float):
        calls.append(t)

    mocker.patch("time.sleep", mocked_sleep)

    sleep = sleep_exponential(
        min_duration=min_duration,
        factor=factor,
        max_duration=max_duration,
        message="Message",
    )

    for i, _ in enumerate(expected):
        sleep(i)

    assert calls == expected


def test_winery_sleep_exponential_negative():
    with pytest.raises(ValueError, match="negative amount"):
        _ = sleep_exponential(
            min_duration=-1, factor=2, max_duration=10, message="Message"
        )


@pytest.mark.shard_max_size(1024)
@pytest.mark.pack_immediately(False)
def test_winery_standalone_packer(shard_max_size, image_pool, postgresql_dsn, storage):
    # create 4 shards
    for i in range(16):
        content = i.to_bytes(256, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    filled = storage.winery.shards_filled
    assert len(filled) == 4

    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert shard_info[shard] == ShardState.FULL

    # Pack a single shard
    assert (
        shard_packer(
            base_dsn=postgresql_dsn,
            shard_max_size=shard_max_size,
            throttle_read=200 * 1024 * 1024,
            throttle_write=200 * 1024 * 1024,
            stop_packing=stop_after_shards(1),
            rbd_pool_name=image_pool.pool_name,
            rbd_map_options=image_pool.map_options,
        )
        == 1
    )

    shard_counts = Counter(state for _, state in storage.winery.base.list_shards())
    assert shard_counts == {
        ShardState.FULL: 3,
        ShardState.PACKED: 1,
    }

    # Clean up the RW shard for the packed one
    assert (
        rw_shard_cleaner(
            base_dsn=postgresql_dsn,
            min_mapped_hosts=0,
            stop_cleaning=stop_after_shards(1),
        )
        == 1
    )

    shard_counts = Counter(state for _, state in storage.winery.base.list_shards())
    assert shard_counts == {
        ShardState.FULL: 3,
        ShardState.READONLY: 1,
    }

    # Pack all remaining shards
    assert (
        shard_packer(
            base_dsn=postgresql_dsn,
            shard_max_size=shard_max_size,
            throttle_read=200 * 1024 * 1024,
            throttle_write=200 * 1024 * 1024,
            stop_packing=stop_after_shards(3),
            rbd_pool_name=image_pool.pool_name,
            rbd_map_options=image_pool.map_options,
        )
        == 3
    )

    shard_counts = Counter(state for _, state in storage.winery.base.list_shards())
    assert shard_counts == {
        ShardState.PACKED: 3,
        ShardState.READONLY: 1,
    }

    # Clean up the RW shard for the packed one
    assert (
        rw_shard_cleaner(
            base_dsn=postgresql_dsn,
            min_mapped_hosts=0,
            stop_cleaning=stop_after_shards(3),
        )
        == 3
    )

    shard_counts = Counter(state for _, state in storage.winery.base.list_shards())
    assert shard_counts == {ShardState.READONLY: 4}


@pytest.mark.shard_max_size(1024)
@pytest.mark.pack_immediately(False)
def test_winery_packer_clean_up_interrupted_shard(
    shard_max_size, image_pool, postgresql_dsn, storage, caplog
):
    caplog.set_level(logging.CRITICAL)

    # create 1 full shard
    for i in range(4):
        content = i.to_bytes(256, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    filled = storage.winery.shards_filled
    assert len(filled) == 1

    shard = filled[0]

    if not image_pool.image_mapped(shard):
        image_pool.image_create(shard)

    with open(image_pool.image_path(shard), "wb") as f:
        f.write(b"SWHShard interrupted bla")

    with caplog.at_level(logging.WARNING, "swh.objstorage.backends.winery.roshard"):
        # Pack a single shard
        ret = shard_packer(
            base_dsn=postgresql_dsn,
            shard_max_size=shard_max_size,
            throttle_read=200 * 1024 * 1024,
            throttle_write=200 * 1024 * 1024,
            stop_packing=stop_after_shards(1),
            rbd_pool_name=image_pool.pool_name,
            rbd_create_images=False,
        )

    assert ret == 1
    found_cleanup_message = False
    found_subprocess_error = False
    for record in caplog.records:
        msg = record.getMessage()
        if image_pool.image_path(shard) in msg:
            if "cleaning it up" in msg:
                found_cleanup_message = True
            elif "failed:" in msg:
                found_subprocess_error = True
    else:
        assert found_cleanup_message and not found_subprocess_error, [
            r.getMessage() for r in caplog.records
        ]


@pytest.mark.shard_max_size(1024)
@pytest.mark.pack_immediately(False)
@pytest.mark.clean_immediately(False)
def test_winery_cli_packer(image_pool, storage, tmp_path, cli_runner):
    # create 4 shards
    for i in range(16):
        content = i.to_bytes(256, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    filled = storage.winery.shards_filled
    assert len(filled) == 4

    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert shard_info[shard] == ShardState.FULL

    with open(tmp_path / "config.yml", "w") as f:
        yaml.safe_dump(
            {"objstorage": {"cls": "winery", **storage.winery.args}}, stream=f
        )

    result = cli_runner.invoke(
        swh_cli_group,
        ("objstorage", "winery", "packer", "--stop-after-shards=4"),
        env={"SWH_CONFIG_FILENAME": str(tmp_path / "config.yml")},
    )

    assert result.exit_code == 0

    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert shard_info[shard] == ShardState.PACKED


@pytest.mark.shard_max_size(1024)
@pytest.mark.pack_immediately(False)
@pytest.mark.clean_immediately(False)
def test_winery_cli_packer_rollback_on_error(image_pool, storage, tmp_path, cli_runner):
    # create 4 shards
    for i in range(16):
        content = i.to_bytes(256, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    filled = storage.winery.shards_filled
    assert len(filled) == 4

    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert shard_info[shard] == ShardState.FULL

    with open(tmp_path / "config.yml", "w") as f:
        yaml.safe_dump(
            {"objstorage": {"cls": "winery", **storage.winery.args}}, stream=f
        )

    # pytest-mock doesn't seem to interact very well with the cli_runner
    def failing_pack(*args, **kwargs):
        raise ValueError("Packing failed")

    orig_pack = swh.objstorage.backends.winery.objstorage.pack
    try:
        swh.objstorage.backends.winery.objstorage.pack = failing_pack
        result = cli_runner.invoke(
            swh_cli_group,
            ("objstorage", "winery", "packer", "--stop-after-shards=4"),
            env={"SWH_CONFIG_FILENAME": str(tmp_path / "config.yml")},
        )
    finally:
        swh.objstorage.backends.winery.objstorage.pack = orig_pack

    assert result.exit_code == 1

    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert (
            shard_info[shard] == ShardState.FULL
        ), f"{shard} in state {shard_info[shard]}"


@pytest.mark.shard_max_size(1024)
@pytest.mark.pack_immediately(False)
def test_winery_cli_rbd(image_pool, storage, tmp_path, cli_runner):
    # create 4 shards
    for i in range(16):
        content = i.to_bytes(256, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    filled = storage.winery.shards_filled
    assert len(filled) == 4

    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert shard_info[shard] == ShardState.FULL

    with open(tmp_path / "config.yml", "w") as f:
        yaml.safe_dump(
            {"objstorage": {"cls": "winery", **storage.winery.args}}, stream=f
        )

    result = cli_runner.invoke(
        swh_cli_group,
        ("objstorage", "winery", "rbd", "--stop-instead-of-waiting"),
        env={"SWH_CONFIG_FILENAME": str(tmp_path / "config.yml")},
    )

    assert result.exit_code == 0

    for shard in filled:
        assert image_pool.image_mapped(shard) == "rw"

    for shard in filled:
        storage.winery.base.set_shard_state(name=shard, new_state=ShardState.PACKED)

    result = cli_runner.invoke(
        swh_cli_group,
        ("objstorage", "winery", "rbd", "--stop-instead-of-waiting"),
        env={"SWH_CONFIG_FILENAME": str(tmp_path / "config.yml")},
    )

    assert result.exit_code == 0

    for shard in filled:
        assert image_pool.image_mapped(shard) == "ro"


@pytest.mark.shard_max_size(1024)
@pytest.mark.pack_immediately(True)
@pytest.mark.clean_immediately(False)
def test_winery_cli_rw_shard_cleaner(
    image_pool, postgresql_dsn, storage, tmp_path, cli_runner
):
    # create 4 shards
    for i in range(16):
        content = i.to_bytes(256, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    filled = storage.winery.shards_filled
    assert len(filled) == 4

    for packer in storage.winery.packers:
        packer.join()
        assert packer.exitcode == 0

    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert shard_info[shard] == ShardState.PACKED

    with open(tmp_path / "config.yml", "w") as f:
        yaml.safe_dump(
            {"objstorage": {"cls": "winery", **storage.winery.args}}, stream=f
        )

    shard_tables = set(storage.winery.base.list_shard_tables())
    for shard in filled:
        assert shard in shard_tables

    result = cli_runner.invoke(
        swh_cli_group,
        ("objstorage", "winery", "rw-shard-cleaner", "--stop-instead-of-waiting"),
        env={"SWH_CONFIG_FILENAME": str(tmp_path / "config.yml")},
    )

    assert result.exit_code == 0

    # No hosts have mapped the shard as remapped, so the cleaner has done nothing
    shard_tables = set(storage.winery.base.list_shard_tables())
    for shard in filled:
        assert shard in shard_tables

    result = cli_runner.invoke(
        swh_cli_group,
        (
            "objstorage",
            "winery",
            "rw-shard-cleaner",
            "--stop-instead-of-waiting",
            "--min-mapped-hosts=0",
        ),
        env={"SWH_CONFIG_FILENAME": str(tmp_path / "config.yml")},
    )

    assert result.exit_code == 0

    # Now we've forced action
    shard_tables = set(storage.winery.base.list_shard_tables())
    for shard in filled:
        assert shard not in shard_tables


@pytest.mark.shard_max_size(1024)
@pytest.mark.pack_immediately(True)
@pytest.mark.clean_immediately(False)
def test_winery_cli_rw_shard_cleaner_rollback_on_error(
    image_pool, postgresql_dsn, storage, tmp_path, cli_runner
):
    # create 4 shards
    for i in range(16):
        content = i.to_bytes(256, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    filled = storage.winery.shards_filled
    assert len(filled) == 4

    for packer in storage.winery.packers:
        packer.join()
        assert packer.exitcode == 0

    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert shard_info[shard] == ShardState.PACKED

    with open(tmp_path / "config.yml", "w") as f:
        yaml.safe_dump(
            {"objstorage": {"cls": "winery", **storage.winery.args}}, stream=f
        )

    shard_tables = set(storage.winery.base.list_shard_tables())
    for shard in filled:
        assert shard in shard_tables

    # pytest-mock doesn't seem to interact very well with the cli_runner
    def failing_cleanup(*args, **kwargs):
        raise ValueError("Cleanup failed")

    orig_cleanup = swh.objstorage.backends.winery.objstorage.cleanup_rw_shard
    try:
        swh.objstorage.backends.winery.objstorage.cleanup_rw_shard = failing_cleanup

        result = cli_runner.invoke(
            swh_cli_group,
            (
                "objstorage",
                "winery",
                "rw-shard-cleaner",
                "--stop-instead-of-waiting",
                "--min-mapped-hosts=0",
            ),
            env={"SWH_CONFIG_FILENAME": str(tmp_path / "config.yml")},
        )
    finally:
        swh.objstorage.backends.winery.objstorage.cleanup_rw_shard = orig_cleanup

    assert result.exit_code == 1

    shard_tables = set(storage.winery.base.list_shard_tables())
    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert shard in shard_tables
        assert shard_info[shard] == ShardState.PACKED


@pytest.mark.shard_max_size(1024)
@pytest.mark.pack_immediately(False)
def test_winery_standalone_packer_never_stop_packing(
    image_pool, postgresql_dsn, shard_max_size, storage
):
    # create 4 shards
    for i in range(16):
        content = i.to_bytes(256, "little")
        obj_id = compute_hash(content, "sha256")
        storage.add(content=content, obj_id=obj_id)

    filled = storage.winery.shards_filled
    assert len(filled) == 4

    shard_info = dict(storage.winery.base.list_shards())
    for shard in filled:
        assert shard_info[shard] == ShardState.FULL

    class NoShardLeft(Exception):
        pass

    called = []

    def wait_five_times(attempt) -> None:
        called.append(attempt)
        if attempt >= 4:
            raise NoShardLeft(attempt)

    with pytest.raises(NoShardLeft):
        shard_packer(
            base_dsn=postgresql_dsn,
            shard_max_size=shard_max_size,
            throttle_read=200 * 1024 * 1024,
            throttle_write=200 * 1024 * 1024,
            wait_for_shard=wait_five_times,
            rbd_pool_name=image_pool.pool_name,
            rbd_map_options=image_pool.map_options,
        )

    assert called == list(range(5))

    shard_counts = Counter(state for _, state in storage.winery.base.list_shards())
    assert shard_counts == {ShardState.PACKED: 4}

    called = []

    with pytest.raises(NoShardLeft):
        rw_shard_cleaner(
            base_dsn=postgresql_dsn,
            min_mapped_hosts=0,
            wait_for_shard=wait_five_times,
        )

    assert called == list(range(5))

    shard_counts = Counter(state for _, state in storage.winery.base.list_shards())
    assert shard_counts == {ShardState.READONLY: 4}


@pytest.mark.shard_max_size(10 * 1024 * 1024)
def test_winery_get_object(winery, image_pool):
    shard = winery.base.locked_shard
    content = b"SOMETHING"
    obj_id = compute_hash(content, "sha256")
    winery.add(content=content, obj_id=obj_id)
    winery.base.set_shard_state(ShardState.FULL)
    winery.base.shard_packing_starts(shard)
    assert pack(shard, **winery.args) is True
    assert winery.get(obj_id) == content


@pytest.mark.skipif("CEPH_HARDCODE_POOL" in os.environ, reason="Ceph pool hardcoded")
def test_winery_ceph_pool(needs_ceph, rbd_map_options):
    name = "IMAGE"
    pool = PoolHelper(
        shard_max_size=10 * 1024 * 1024,
        rbd_pool_name="test-winery-ceph-pool",
        rbd_map_options=rbd_map_options,
    )
    pool.remove()
    pool.pool_create()
    assert pool.image_mapped(name) is None
    pool.image_create(name)
    assert pool.image_mapped(name) == "rw"
    p = pool.image_path(name)
    assert p.endswith(name)
    something = "SOMETHING"
    open(p, "w").write(something)
    assert open(p).read(len(something)) == something
    assert pool.image_list() == [name]
    pool.image_remap_ro(name)
    if pool.image_mapped(name) == "rw":
        raise ValueError(
            "Remapping image read-only kept write permissions. "
            "Are the udev rules properly installed?"
        )
    assert pool.image_mapped(name) == "ro"
    pool.images_remove()
    assert pool.image_list() == []
    assert pool.image_mapped(name) is None
    pool.remove()
    assert pool.image_list() == []


@pytest.mark.shard_max_size(10 * 1024 * 1024)
def test_winery_bench_work_ro_rw(storage, image_pool, tmpdir):
    #
    # rw worker creates a shard
    #
    locked_shard = storage.winery.base.locked_shard
    shards_info = list(storage.winery.base.list_shards())
    assert shards_info == [(locked_shard, ShardState.WRITING)]
    assert (
        work("rw", storage=storage, time_remaining=datetime.timedelta(seconds=300))
        == "rw"
    )
    shards_info = dict(storage.winery.base.list_shards())
    assert shards_info[locked_shard].image_available
    #
    # ro worker reads a shard
    #
    args = {**storage.winery.args, "readonly": True}
    assert (
        work(
            "ro",
            storage=args,
            worker_args={"ro": {"max_request": 1}},
            time_remaining=datetime.timedelta(seconds=300),
        )
        == "ro"
    )


@pytest.mark.shard_max_size(10 * 1024 * 1024)
def test_winery_bench_work_pack(storage, image_pool):
    pack_args = {
        "base_dsn": storage.winery.args["base_dsn"],
        "shard_max_size": storage.winery.args["shard_max_size"],
        "throttle_read": storage.winery.args["throttle_read"],
        "throttle_write": storage.winery.args["throttle_write"],
        "rbd_pool_name": image_pool.pool_name,
    }
    assert (
        work(
            "pack",
            storage=storage,
            worker_args={"pack": pack_args},
            time_remaining=datetime.timedelta(seconds=300),
        )
        == "pack"
    )


@pytest.mark.shard_max_size(10 * 1024 * 1024)
def test_winery_bench_work_rbd(storage, image_pool):
    rbd_args = {
        "base_dsn": storage.winery.args["base_dsn"],
        "shard_max_size": storage.winery.args["shard_max_size"],
        "duration": 1,
        "rbd_pool_name": image_pool.pool_name,
        "rbd_map_options": image_pool.map_options,
    }
    assert (
        work(
            "rbd",
            storage=storage,
            worker_args={"rbd": rbd_args},
            time_remaining=datetime.timedelta(seconds=300),
        )
        == "rbd"
    )


@pytest.mark.shard_max_size(10 * 1024 * 1024)
def test_winery_bench_work_rw_shard_cleaner(storage):
    rw_shard_cleaner_args = {
        "base_dsn": storage.winery.args["base_dsn"],
    }
    assert (
        work(
            "rw_shard_cleaner",
            storage=storage,
            worker_args={"rw_shard_cleaner": rw_shard_cleaner_args},
            time_remaining=datetime.timedelta(seconds=300),
        )
        == "rw_shard_cleaner"
    )


@pytest.mark.shard_max_size(10 * 1024 * 1024)
@pytest.mark.pack_immediately(False)
def test_winery_bench_rw_object_limit(storage):
    object_limit = 15
    worker = RWWorker(
        storage, object_limit=object_limit, single_shard=False, block_until_packed=False
    )

    assert worker.run(time_remaining=datetime.timedelta(seconds=300)) == "rw"

    with storage.winery.base.pool.connection() as db:
        c = db.execute("SELECT count(*) from signature2shard")
        assert c.fetchone() == (object_limit,)


@pytest.mark.shard_max_size(10 * 1024 * 1024)
@pytest.mark.pack_immediately(True)
def test_winery_bench_rw_block_until_packed(storage, image_pool):
    worker = RWWorker(storage, single_shard=True, block_until_packed=False)

    assert worker.run(time_remaining=datetime.timedelta(seconds=300)) == "rw"

    packed = 0
    for packer in storage.winery.packers:
        packer.join()
        assert packer.exitcode == 0
        packed += 1

    assert packed > 0, "did not have any packers to wait for"


@pytest.mark.shard_max_size(1024 * 1024)
@pytest.mark.pack_immediately(True)
def test_winery_bench_rw_block_until_packed_multiple_shards(storage, image_pool):
    # 1000 objects will create multiple shards when the limit is 1MB
    worker = RWWorker(
        storage, object_limit=1000, single_shard=False, block_until_packed=False
    )

    assert worker.run(time_remaining=datetime.timedelta(seconds=300)) == "rw"

    packed = 0
    for packer in storage.winery.packers:
        packer.join()
        assert packer.exitcode == 0
        packed += 1

    assert packed > 0, "did not have any packers to wait for"


@dataclass
class WineryBenchOptions:
    storage_config: Dict[str, Any]
    workers_per_kind: Dict[WorkerKind, int]
    worker_args: Dict[WorkerKind, Dict]
    duration: float


@pytest.fixture
def bench_options(pytestconfig, postgresql_dsn, rbd_map_options) -> WineryBenchOptions:
    output_dir = pytestconfig.getoption("--winery-bench-output-directory")
    shard_max_size = pytestconfig.getoption("--winery-bench-shard-max-size")
    pack_immediately = pytestconfig.getoption("--winery-bench-pack-immediately")
    duration = pytestconfig.getoption("--winery-bench-duration")

    storage_config = {
        "output_dir": output_dir,
        "shard_max_size": shard_max_size,
        "pack_immediately": pack_immediately,
        "base_dsn": postgresql_dsn,
        "throttle_read": pytestconfig.getoption("--winery-bench-throttle-read"),
        "throttle_write": pytestconfig.getoption("--winery-bench-throttle-write"),
        "rbd_pool_name": pytestconfig.getoption("--winery-bench-rbd-pool"),
    }
    workers_per_kind: Dict[WorkerKind, int] = {
        "ro": pytestconfig.getoption("--winery-bench-ro-workers"),
        "rw": pytestconfig.getoption("--winery-bench-rw-workers"),
        "stats": (
            1 if pytestconfig.getoption("--winery-bench-stats-interval") > 0 else 0
        ),
    }
    worker_args: Dict[WorkerKind, Dict] = {
        "ro": {
            "max_request": pytestconfig.getoption(
                "--winery-bench-ro-worker-max-request"
            )
        },
        "pack": {
            "base_dsn": postgresql_dsn,
            "output_dir": output_dir,
            "shard_max_size": shard_max_size,
            "rbd_create_images": False,
            "rbd_pool_name": pytestconfig.getoption("--winery-bench-rbd-pool"),
            "throttle_read": pytestconfig.getoption("--winery-bench-throttle-read"),
            "throttle_write": pytestconfig.getoption("--winery-bench-throttle-write"),
        },
        "stats": {
            "base_dsn": postgresql_dsn,
            "shard_max_size": shard_max_size,
            "interval": pytestconfig.getoption("--winery-bench-stats-interval"),
        },
        "rw_shard_cleaner": {
            "base_dsn": postgresql_dsn,
        },
        "rbd": {
            "base_dsn": postgresql_dsn,
            "shard_max_size": shard_max_size,
            "rbd_pool_name": pytestconfig.getoption("--winery-bench-rbd-pool"),
            "rbd_map_options": rbd_map_options,
            "duration": duration,
        },
    }

    if not pack_immediately:
        worker_args["rw"] = {"block_until_packed": False}
        workers_per_kind["pack"] = pytestconfig.getoption("--winery-bench-pack-workers")
        workers_per_kind["rw_shard_cleaner"] = 1
        workers_per_kind["rbd"] = 1

    return WineryBenchOptions(
        storage_config,
        workers_per_kind,
        worker_args,
        duration,
    )


# Only run this one on a Ceph image pool as we won’t be running “real”
# benchmarks using the file-backed backend.
@pytest.mark.use_benchmark_flags
def test_winery_bench_real(bench_options, ceph_pool):
    count = call_async(Bench(**asdict(bench_options)).run)
    assert count > 0


@pytest.mark.use_benchmark_flags
def test_winery_bench_fake(bench_options, mocker):
    class _ROWorker(ROWorker):
        def run(self, time_remaining: datetime.timedelta):
            logger.info(
                "running ro for %s, time remaining: %s",
                bench_options.duration,
                time_remaining,
            )
            return "ro"

    class _RWWorker(RWWorker):
        def run(self, time_remaining: datetime.timedelta):
            logger.info(
                "running rw for %s, time remaining: %s",
                bench_options.duration,
                time_remaining,
            )
            return "rw"

    class _PackWorker(PackWorker):
        def run(self):
            logger.info("running pack for %s", bench_options.duration)
            return "pack"

    class _RBDWorker(RBDWorker):
        def run(self):
            logger.info("running rbd for %s", bench_options.duration)
            return "rbd"

    class _RWShardCleanerWorker(RWShardCleanerWorker):
        def run(self):
            logger.info("running rw_shard_cleaner for %s", bench_options.duration)
            return "rw_shard_cleaner"

    class _StatsPrinter(StatsPrinter):
        def run(self, time_remaining: datetime.timedelta):
            logger.info(
                "running stats for %s, remaining: %s",
                bench_options.duration,
                time_remaining,
            )
            return "stats"

    mocker.patch("swh.objstorage.tests.winery_benchmark.ROWorker", _ROWorker)
    mocker.patch("swh.objstorage.tests.winery_benchmark.RWWorker", _RWWorker)
    mocker.patch("swh.objstorage.tests.winery_benchmark.PackWorker", _PackWorker)
    mocker.patch("swh.objstorage.tests.winery_benchmark.RBDWorker", _RBDWorker)
    mocker.patch(
        "swh.objstorage.tests.winery_benchmark.RWShardCleanerWorker",
        _RWShardCleanerWorker,
    )
    mocker.patch("swh.objstorage.tests.winery_benchmark.StatsPrinter", _StatsPrinter)
    mocker.patch(
        "swh.objstorage.tests.winery_benchmark.Bench.timeout", side_effect=lambda: True
    )

    count = call_async(Bench(**asdict(bench_options)).run)
    assert count == sum(bench_options.workers_per_kind.values())


def test_winery_leaky_bucket_tick(mocker):
    total = 100
    half = 50
    b = LeakyBucket(total)
    sleep = mocker.spy(time, "sleep")
    assert b.current == b.total
    sleep.assert_not_called()
    #
    # Bucket is at 100, add(50) => drops to 50
    #
    b.add(half)
    assert b.current == half
    sleep.assert_not_called()
    #
    # Bucket is at 50, add(50) => drops to 0
    #
    b.add(half)
    assert b.current == 0
    sleep.assert_not_called()
    #
    # Bucket is at 0, add(50) => waits until it is at 50 and then drops to 0
    #
    b.add(half)
    assert b.current == 0
    sleep.assert_called_once()
    #
    # Sleep more than one second, bucket is full again, i.e. at 100
    #
    time.sleep(2)
    mocker.resetall()
    b.add(0)
    assert b.current == total
    sleep.assert_not_called()
    #
    # Bucket is full at 100 and and waits when requesting 150 which is
    # more than it can contain
    #
    b.add(total + half)
    assert b.current == 0
    sleep.assert_called_once()
    mocker.resetall()
    #
    # Bucket is empty and and waits when requesting 150 which is more
    # than it can contain
    #
    b.add(total + half)
    assert b.current == 0
    sleep.assert_called_once()
    mocker.resetall()


def test_winery_leaky_bucket_reset():
    b = LeakyBucket(100)
    assert b.total == 100
    assert b.current == b.total
    b.reset(50)
    assert b.total == 50
    assert b.current == b.total
    b.reset(100)
    assert b.total == 100
    assert b.current == 50


def test_winery_bandwidth_calculator(mocker):
    now = 1

    def monotonic():
        return now

    mocker.patch("time.monotonic", side_effect=monotonic)
    b = BandwidthCalculator()
    assert b.get() == 0
    count = 100 * 1024 * 1024
    going_up = []
    for t in range(b.duration):
        now += 1
        b.add(count)
        going_up.append(b.get())
    assert b.get() == count
    going_down = []
    for t in range(b.duration - 1):
        now += 1
        b.add(0)
        going_down.append(b.get())
    going_down.reverse()
    assert going_up[:-1] == going_down
    assert len(b.history) == b.duration - 1


def test_winery_io_throttler(postgresql_dsn, mocker):
    sleep = mocker.spy(time, "sleep")
    speed = 100
    i = IOThrottler("read", base_dsn=postgresql_dsn, throttle_read=100)
    count = speed
    i.add(count)
    sleep.assert_not_called()
    i.add(count)
    sleep.assert_called_once()
    #
    # Force slow down
    #
    mocker.resetall()
    i.sync_interval = 0
    i.max_speed = 1
    assert i.max_speed != i.bucket.total
    i.add(2)
    assert i.max_speed == i.bucket.total
    sleep.assert_called_once()


def test_winery_throttler(postgresql_dsn):
    t = Throttler(base_dsn=postgresql_dsn, throttle_read=100, throttle_write=100)

    base = {}
    key = "KEY"
    content = "CONTENT"

    def reader(k):
        return base[k]

    def writer(k, v):
        base[k] = v
        return True

    assert t.throttle_add(writer, key, content) is True
    assert t.throttle_get(reader, key) == content


def test_winery_stats(tmpdir):
    s = Stats(None)
    assert s.stats_active is False
    s = Stats(tmpdir / "stats")
    assert s.stats_active is True
    assert os.path.exists(s.stats_filename)
    size = os.path.getsize(s.stats_filename)
    s._stats_flush_interval = 0
    k = "KEY"
    v = "CONTENT"
    s.stats_read(k, v)
    s.stats_write(k, v)
    s.stats_read(k, v)
    s.stats_write(k, v)
    s.__del__()
    assert os.path.getsize(s.stats_filename) > size



# Copyright (C) 2022-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import collections
import datetime
import logging
import time

from .database import Database

logger = logging.getLogger(__name__)


class LeakyBucket:
    """
    Leaky bucket that can contain at most `total` and leaks it within a second.
    If adding (with the add method) more than `total` per second, it will sleep
    until the bucket can be filled without overflowing.

    The capacity of the bucket can be changed dynamically with the reset method.
    If the new capacity is lower than it previously was, the overflow is ignored.
    """

    def __init__(self, total):
        self.updated = 0.0
        self.current = 0.0
        self.reset(total)

    def reset(self, total):
        self.total = total
        self.current = min(self.total, self.current)
        self._tick()

    def add(self, count):
        self._tick()
        if count > self.current:
            time.sleep((count - self.current) / self.total)
            self._tick()
        self.current -= min(self.total, count)

    def _tick(self):
        now = time.monotonic()
        self.current += self.total * (now - self.updated)
        self.current = int(min(self.total, self.current))
        self.updated = now


class BandwidthCalculator:
    """Keeps a histogram (of length `duration`, defaults to 60) where
    each element is the number of bytes read or written within a
    second.

    Only the last `duration` seconds are represented in the histogram:
    after each second the oldest element is discarded.

    The `add` method is called to add a value to the current second.

    The `get` method retrieves the current bandwidth usage which is
    the average of all values in the histogram.
    """

    def __init__(self):
        self.duration = 60
        self.history = collections.deque([0] * (self.duration - 1), self.duration - 1)
        self.current = 0
        self.current_second = 0

    def add(self, count):
        current_second = int(time.monotonic())
        if current_second > self.current_second:
            self.history.append(self.current)
            self.history.extend(
                [0] * min(self.duration, current_second - self.current_second - 1)
            )
            self.current_second = current_second
            self.current = 0
        self.current += count

    def get(self):
        return (sum(self.history) + self.current) / self.duration


class IOThrottler(Database):
    """Throttle IO (either read or write, depending on the `name`
    argument). The maximum speed in bytes is from the throttle_`name`
    argument and controlled by a LeakyBucket that guarantees it won't
    go any faster.

    Every `sync_interval` seconds the current bandwidth reported by
    the BandwidthCalculator instance is written into a row in a table
    shared with other instances of IOThrottler. The cumulated
    bandwidth of all other instances is retrieved from the same table.
    If it exceeds `max_speed`, the LeakyBucket instance is reset to
    only allow max_speed/(number of instances) so that the total
    bandwidth is shared equally between instances.
    """

    def __init__(self, name, **kwargs):
        super().__init__(
            dsn=kwargs["base_dsn"],
            application_name=kwargs.get("application_name", "SWH Winery Throttler"),
        )
        self.name = name
        self.init_db()
        self.last_sync = 0
        self.max_speed = kwargs["throttle_" + name]
        self.bucket = LeakyBucket(self.max_speed)
        self.bandwidth = BandwidthCalculator()

    def init_db(self):
        with self.pool.connection() as db, db.transaction(), db.cursor() as cur:
            cur.execute(
                f"INSERT INTO t_{self.name} (updated, bytes) VALUES (%s, %s) RETURNING id",
                (datetime.datetime.now(), 0),
            )
            self.rowid = cur.fetchone()[0]
            cur.execute(
                f"SELECT * FROM t_{self.name} WHERE id = %s FOR UPDATE", (self.rowid,)
            )
            cur.execute(
                f"DELETE FROM t_{self.name} WHERE id IN ("
                f"SELECT id FROM t_{self.name} WHERE updated < NOW() - INTERVAL '30 days' "
                " FOR UPDATE SKIP LOCKED)"
            )
        self.sync_interval = 60

    def download_info(self):
        with self.pool.connection() as db:
            cur = db.execute(
                f"SELECT COUNT(*), SUM(bytes) FROM t_{self.name} "
                "WHERE bytes > 0 AND updated > NOW() - INTERVAL '5 minutes'"
            )
            return cur.fetchone()

    def upload_info(self):
        bytes = int(self.bandwidth.get())
        logger.debug("%d: upload %s/s", self.rowid, bytes)

        with self.pool.connection() as db:
            db.execute(
                f"UPDATE t_{self.name} SET updated = %s, bytes = %s WHERE id = %s",
                (datetime.datetime.now(), bytes, self.rowid),
            )

    def add(self, count):
        self.bucket.add(count)
        self.bandwidth.add(count)
        self.maybe_sync()

    def sync(self):
        self.upload_info()
        (others_count, total_usage) = self.download_info()
        logger.debug(
            "%d: sync others_count=%s total_usage=%s",
            self.rowid,
            others_count,
            total_usage,
        )
        if others_count > 0 and total_usage > self.max_speed:
            self.bucket.reset(self.max_speed / others_count)

    def maybe_sync(self):
        now = time.monotonic()
        if now - self.last_sync > self.sync_interval:
            self.sync()
            self.last_sync = now


class Throttler:
    """Throttle reads and writes to not exceed limits imposed by the
    `thottle_read` and `throttle_write` arguments, as measured by the
    cumulated bandwidth reported by each Throttler instance.
    """

    def __init__(self, **kwargs):
        self.read = IOThrottler("read", **kwargs)
        self.write = IOThrottler("write", **kwargs)

    def throttle_get(self, fun, key):
        content = fun(key)
        self.read.add(len(content))
        return content

    def throttle_add(self, fun, obj_id, content):
        self.write.add(len(obj_id) + len(content))
        return fun(obj_id, content)



# Copyright (C) 2016-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import importlib

from swh.objstorage.interface import ObjStorageInterface
from swh.objstorage.multiplexer import MultiplexerObjStorage
from swh.objstorage.multiplexer.filter import add_filters
from swh.objstorage.objstorage import ObjStorage

__all__ = ["get_objstorage", "ObjStorage"]


OBJSTORAGE_IMPLEMENTATIONS = {
    "pathslicing": ".backends.pathslicing.PathSlicingObjStorage",
    "remote": ".api.client.RemoteObjStorage",
    "memory": ".backends.in_memory.InMemoryObjStorage",
    "seaweedfs": ".backends.seaweedfs.SeaweedFilerObjStorage",
    "random": ".backends.generator.RandomGeneratorObjStorage",
    "http": ".backends.http.HTTPReadOnlyObjStorage",
    "noop": ".backends.noop.NoopObjStorage",
    "azure": ".backends.azure.AzureCloudObjStorage",
    "azure-prefixed": ".backends.azure.PrefixedAzureCloudObjStorage",
    "s3": ".backends.libcloud.AwsCloudObjStorage",
    "swift": ".backends.libcloud.OpenStackCloudObjStorage",
    "winery": ".backends.winery.WineryObjStorage",
}


def get_objstorage(cls: str, **kwargs) -> ObjStorageInterface:
    """Create an ObjStorage using the given implementation class.

    Args:
        cls: objstorage class unique key contained in the
            _STORAGE_CLASSES dict.
        kwargs: arguments for the required class of objstorage
                that must match exactly the one in the `__init__` method of the
                class.
    Returns:
        subclass of ObjStorage that match the given `storage_class` argument.
    Raises:
        ValueError: if the given storage class is not a valid objstorage
            key.
    """
    class_path = OBJSTORAGE_IMPLEMENTATIONS.get(cls)
    if class_path is None:
        raise ValueError(
            "Unknown storage class `%s`. Supported: %s"
            % (cls, ", ".join(OBJSTORAGE_IMPLEMENTATIONS))
        )

    if "." in class_path:
        (module_path, class_name) = class_path.rsplit(".", 1)
        try:
            module = importlib.import_module(module_path, package=__package__)
        except ImportError as e:
            raise ValueError(f"Storage class {cls} is not available: {e.args[0]}")
        ObjStorage = getattr(module, class_name)
    else:
        ObjStorage = globals()[class_path]

    return ObjStorage(**kwargs)


def _construct_filtered_objstorage(storage_conf, filters_conf):
    return add_filters(get_objstorage(**storage_conf), filters_conf)


OBJSTORAGE_IMPLEMENTATIONS["filtered"] = "_construct_filtered_objstorage"


def _construct_multiplexer_objstorage(objstorages):
    storages = [get_objstorage(**conf) for conf in objstorages]
    return MultiplexerObjStorage(storages)


OBJSTORAGE_IMPLEMENTATIONS["multiplexer"] = "_construct_multiplexer_objstorage"



# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import sys

from swh.core.api.gunicorn_config import *  # noqa


def worker_int(worker):
    """Run on_shutdown callback for storage when a worker is terminating"""
    if "swh.objstorage.api.server" not in sys.modules:
        return

    objstorage = sys.modules["swh.objstorage.api.server"].objstorage

    if on_shutdown := getattr(objstorage, "on_shutdown", None):
        on_shutdown()



# Copyright (C) 2021-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
import logging

from psycopg_pool import ConnectionPool

logger = logging.getLogger(__name__)


class Database(abc.ABC):
    def __init__(self, dsn: str, application_name: str):
        self.dsn = dsn
        self.application_name = application_name
        self._pool = None

    @property
    def pool(self):
        if not self._pool:
            self._pool = ConnectionPool(
                conninfo=self.dsn,
                kwargs={
                    "application_name": self.application_name,
                    "fallback_application_name": "SWH Winery",
                    "autocommit": True,
                },
                name=f"pool-{self.application_name}"
                if self.application_name
                else "pool",
                min_size=0,
                max_size=4,
                open=True,
                max_idle=5,
                check=ConnectionPool.check_connection,
            )
        return self._pool

    def list_shard_tables(self):
        QUERY = """
        SELECT c.relname
        FROM pg_catalog.pg_class c
             LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
             LEFT JOIN pg_catalog.pg_am am ON am.oid = c.relam
        WHERE c.relkind = 'r'
              AND n.nspname <> 'pg_catalog'
              AND n.nspname !~ '^pg_toast'
              AND n.nspname <> 'information_schema'
              AND pg_catalog.pg_table_is_visible(c.oid)
              AND c.relname ~ '^shard_'
              AND c.relname <> 'shard_template'
        """
        with self.pool.connection() as db:
            c = db.execute(QUERY)
            return [r[0].removeprefix("shard_") for r in c]

    def __del__(self):
        # Release the connection pool
        if self._pool:
            self._pool.close()
            self._pool = None



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import random
from string import ascii_lowercase

import pytest

from swh.objstorage.exc import ObjCorruptedError, ObjNotFoundError
from swh.objstorage.factory import get_objstorage
from swh.objstorage.multiplexer.filter import read_only
from swh.objstorage.objstorage import compute_hash


def get_random_content():
    return bytes("".join(random.sample(ascii_lowercase, 10)), "utf8")


class ReadOnlyFilterTestCase:
    # Read only filter should not allow writing

    @pytest.fixture(autouse=True)
    def objstorage(self, tmpdir):
        pstorage = {
            "cls": "pathslicing",
            "root": tmpdir,
            "slicing": "0:5",
        }
        base_storage = get_objstorage(**pstorage)
        self.storage = get_objstorage(
            "filtered", storage_conf=pstorage, filters_conf=[read_only()]
        )
        self.valid_content = b"pre-existing content"
        self.invalid_content = b"invalid_content"
        self.true_invalid_content = b"Anything that is not correct"
        self.absent_content = b"non-existent content"
        # Create a valid content.
        self.valid_id = compute_hash(self.valid_content)
        base_storage.add(self.valid_content, obj_id=self.valid_id)
        # Create an invalid id and add a content with it.
        self.invalid_id = compute_hash(self.true_invalid_content)
        base_storage.add(self.invalid_content, obj_id=self.invalid_id)
        # Compute an id for a non-existing content.
        self.absent_id = compute_hash(self.absent_content)

    def test_can_contains(self):
        assert self.valid_id in self.storage
        assert self.invalid_id in self.storage
        assert self.absent_id not in self.storage

    def test_can_iter(self):
        assert {"sha1": self.valid_id} in iter(self.storage)
        assert {"sha1": self.invalid_id} in iter(self.storage)

    def test_can_len(self):
        assert len(self.storage) == 2

    def test_can_get(self):
        assert self.valid_content == self.storage.get(self.valid_id)
        assert self.invalid_content == self.storage.get(self.invalid_id)

    def test_can_check(self):
        with pytest.raises(ObjNotFoundError):
            self.storage.check(self.absent_id)
        with pytest.raises(ObjCorruptedError):
            self.storage.check(self.invalid_id)
        self.storage.check(self.valid_id)

    def test_cannot_add(self):
        new_id = self.storage.add(b"New content")
        result = self.storage.add(self.valid_content, self.valid_id)
        assert new_id is None
        assert result is None

    def test_cannot_restore(self):
        result = self.storage.restore(self.valid_content, self.valid_id)
        assert result is None



from itertools import count, islice, repeat
import logging
import random
from typing import Generator, Iterator, Optional, cast

from swh.objstorage.constants import ID_HASH_ALGO
from swh.objstorage.interface import CompositeObjId, ObjId
from swh.objstorage.objstorage import DEFAULT_LIMIT, ObjStorage

logger = logging.getLogger(__name__)


class Randomizer:
    def __init__(self):
        self.size = 0
        self.read(1024)  # create a not-so-small initial buffer

    def read(self, size):
        if size >= self.size:
            with open("/dev/urandom", "rb") as fobj:
                self.data = fobj.read(2 * size)
                self.size = len(self.data)
        # pick a random subset of our existing buffer
        idx = random.randint(0, self.size - size - 1)
        return self.data[idx : idx + size]


def gen_sizes():
    """generates numbers according to the rought distribution of file size in the
    SWH archive
    """
    # these are the histogram bounds of the pg content.length column
    bounds = [
        0,
        2,
        72,
        119,
        165,
        208,
        256,
        300,
        345,
        383,
        429,
        474,
        521,
        572,
        618,
        676,
        726,
        779,
        830,
        879,
        931,
        992,
        1054,
        1119,
        1183,
        1244,
        1302,
        1370,
        1437,
        1504,
        1576,
        1652,
        1725,
        1806,
        1883,
        1968,
        2045,
        2133,
        2236,
        2338,
        2433,
        2552,
        2659,
        2774,
        2905,
        3049,
        3190,
        3322,
        3489,
        3667,
        3834,
        4013,
        4217,
        4361,
        4562,
        4779,
        5008,
        5233,
        5502,
        5788,
        6088,
        6396,
        6728,
        7094,
        7457,
        7835,
        8244,
        8758,
        9233,
        9757,
        10313,
        10981,
        11693,
        12391,
        13237,
        14048,
        14932,
        15846,
        16842,
        18051,
        19487,
        20949,
        22595,
        24337,
        26590,
        28840,
        31604,
        34653,
        37982,
        41964,
        46260,
        51808,
        58561,
        66584,
        78645,
        95743,
        122883,
        167016,
        236108,
        421057,
        1047367,
        55056238,
    ]

    nbounds = len(bounds)
    for i in count():
        idx = random.randint(1, nbounds - 1)
        lower = bounds[idx - 1]
        upper = bounds[idx]
        yield random.randint(lower, upper - 1)


def gen_random_content(total=None, filesize=None):
    """generates random (file) content which sizes roughly follows the SWH
    archive file size distribution (by default).

    Args:
        total (int): the total number of objects to generate. Infinite if
            unset.
        filesize (int): generate objects with fixed size instead of random
            ones.

    """
    randomizer = Randomizer()
    if filesize:
        gen = repeat(filesize)
    else:
        gen = gen_sizes()
    if total:
        gen = islice(gen, total)
    for objsize in gen:
        yield randomizer.read(objsize)


class RandomGeneratorObjStorage(ObjStorage):
    """A stupid read-only storage that generates blobs for testing purpose."""

    def __init__(self, filesize=None, total=None, **kwargs):
        super().__init__()
        if filesize:
            filesize = int(filesize)
        self.filesize = filesize
        if total:
            total = int(total)
        self.total = total
        self._content_generator = None

    @property
    def content_generator(self):
        if self._content_generator is None:
            self._content_generator = gen_random_content(self.total, self.filesize)
        return self._content_generator

    def check_config(self, *, check_write):
        return True

    def __contains__(self, obj_id, *args, **kwargs):
        return False

    def __iter__(self) -> Iterator[CompositeObjId]:
        i = 1
        while True:
            j = yield {ID_HASH_ALGO: b"%d" % i}
            if self.total and i >= self.total:
                logger.debug("DONE")
                break
            if j is not None:
                i = j
            else:
                i += 1

    def get(self, obj_id, *args, **kwargs):
        return next(self.content_generator)

    def add(self, content, obj_id, check_presence=True, *args, **kwargs):
        pass

    def check(self, obj_id, *args, **kwargs):
        return True

    def delete(self, obj_id, *args, **kwargs):
        return True

    def list_content(
        self,
        last_obj_id: Optional[ObjId] = None,
        limit: Optional[int] = DEFAULT_LIMIT,
    ) -> Iterator[CompositeObjId]:
        if isinstance(last_obj_id, dict):
            last_obj_id = last_obj_id[ID_HASH_ALGO]
        it = cast(Generator[CompositeObjId, int, None], iter(self))
        if last_obj_id:
            next(it)
            it.send(int(last_obj_id))
        return islice(it, limit)



# Copyright (C) 2019-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from typing import Iterator
from urllib.parse import urljoin, urlparse

import requests

from swh.objstorage.objstorage import DEFAULT_LIMIT

LOGGER = logging.getLogger(__name__)


class HttpFiler:
    """Simple class that encapsulates access to a seaweedfs filer service.

    TODO: handle errors
    """

    def __init__(self, url, pool_maxsize=100):
        if not url.endswith("/"):
            url = url + "/"
        self.url = url
        self.baseurl = urljoin(url, "/")
        self.basepath = urlparse(url).path

        self.session = requests.Session()
        self.session.headers["Accept"] = "application/json"
        adapter = requests.adapters.HTTPAdapter(
            pool_connections=pool_maxsize, pool_maxsize=pool_maxsize
        )
        self.session.mount(self.url, adapter)
        self.batchsize = DEFAULT_LIMIT

    def build_url(self, path):
        assert path == self.basepath or path.startswith(self.basepath)
        return urljoin(self.baseurl, path)

    def get(self, remote_path):
        url = self.build_url(remote_path)
        LOGGER.debug("Get file %s", url)
        resp = self.session.get(url)
        resp.raise_for_status()
        return resp.content

    def exists(self, remote_path):
        url = self.build_url(remote_path)
        LOGGER.debug("Check file %s", url)
        return self.session.head(url).status_code == 200

    def put(self, fp, remote_path):
        url = self.build_url(remote_path)
        LOGGER.debug("Put file %s", url)
        return self.session.post(url, files={"file": fp})

    def delete(self, remote_path):
        url = self.build_url(remote_path)
        LOGGER.debug("Delete file %s", url)
        return self.session.delete(url)

    def iterfiles(self, dir: str, last_file_name: str = "") -> Iterator[str]:
        """Recursively yield absolute file names

        Args:
            dir: retrieve file names starting from this directory; must
                be an absolute path.
            last_file_name: if given, starts from the file just after; must
                be basename.

        Yields:
            absolute file names

        """
        if not dir.endswith("/"):
            dir = dir + "/"
        # first, generates files going "down" the tree from current position
        # (last_file_name)
        yield from self._iter_files(dir, last_file_name)

        # then, continue iterate going up the tree
        while dir != self.basepath:
            dir, last = dir[:-1].rsplit("/", 1)
            dir += "/"
            yield from self._iter_files(dir, last_file_name=last)

    def _iter_files(self, dir: str, last_file_name: str = "") -> Iterator[str]:
        for entry in self._iter_one_dir(dir, last_file_name):
            fullpath = entry["FullPath"]
            if entry["Mode"] & 1 << 31:  # it's a directory, recurse
                # see https://pkg.go.dev/io/fs#FileMode
                yield from self._iter_files(fullpath)
            else:
                yield fullpath

    def _iter_one_dir(self, remote_path: str, last_file_name: str = ""):
        url = self.build_url(remote_path)
        params = {"limit": self.batchsize}
        if last_file_name:
            params["lastFileName"] = last_file_name

        LOGGER.debug("List directory %s", url)
        while True:
            rsp = self.session.get(url, params=params)
            if rsp.ok:
                dircontent = rsp.json()
                if dircontent["Entries"]:
                    yield from dircontent["Entries"]
                if not dircontent["ShouldDisplayLoadMore"]:
                    break
                params["lastFileName"] = dircontent["LastFileName"]

            else:
                LOGGER.error('Error listing "%s". [HTTP %d]', url, rsp.status_code)
                break



# Copyright (C) 2015-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import inspect
from typing import Tuple

import pytest
import requests

from swh.objstorage.exc import ObjCorruptedError, ObjNotFoundError
from swh.objstorage.interface import CompositeObjId, ObjStorageInterface
from swh.objstorage.objstorage import compute_hash, decompressors


class ObjStorageTestFixture:
    num_objects = 1200

    @pytest.fixture(autouse=True)
    def objstorage(self, swh_objstorage):
        self.storage = swh_objstorage

    def fill_objstorage(self, num_objects):
        all_ids = []
        for i in range(num_objects):
            content = b"content %d" % i
            obj_id = compute_hash(content)
            self.storage.add(content, obj_id, check_presence=False)
            all_ids.append({"sha1": obj_id})
        all_ids.sort(key=lambda d: d["sha1"])
        return all_ids

    def test_types(self):
        """Checks all methods of ObjStorageInterface are implemented by this
        backend, and that they have the same signature."""
        # Create an instance of the protocol (which cannot be instantiated
        # directly, so this creates a subclass, then instantiates it)
        interface = type("_", (ObjStorageInterface,), {})()

        assert "get_batch" in dir(interface)

        missing_methods = []

        for meth_name in dir(interface):
            if meth_name.startswith("_") and meth_name not in (
                "__iter__",
                "__contains__",
            ):
                continue
            interface_meth = getattr(interface, meth_name)
            concrete_meth = getattr(self.storage, meth_name)

            expected_signature = inspect.signature(interface_meth)
            actual_signature = inspect.signature(concrete_meth)

            assert expected_signature == actual_signature, meth_name

        assert missing_methods == []

        # If all the assertions above succeed, then this one should too.
        # But there's no harm in double-checking.
        # And we could replace the assertions above by this one, but unlike
        # the assertions above, it doesn't explain what is missing.
        assert isinstance(self.storage, ObjStorageInterface)

    def hash_content(self, content):
        obj_id = compute_hash(content)
        return content, obj_id

    def compositehash_content(self, content) -> Tuple[bytes, CompositeObjId]:
        obj_id = compute_hash(content)
        return content, {"sha1": obj_id}

    def assertContentMatch(self, obj_id, expected_content):  # noqa
        content = self.storage.get(obj_id)
        assert content == expected_content

    def test_check_config(self):
        assert self.storage.check_config(check_write=False)
        assert self.storage.check_config(check_write=True)

    def test_contains(self):
        content_p, obj_id_p = self.hash_content(b"contains_present")
        content_m, obj_id_m = self.hash_content(b"contains_missing")
        self.storage.add(content_p, obj_id=obj_id_p)
        assert obj_id_p in self.storage
        assert obj_id_m not in self.storage

    def test_contains_composite(self):
        content_p, obj_id_p = self.compositehash_content(b"contains_present")
        content_m, obj_id_m = self.compositehash_content(b"contains_missing")
        self.storage.add(content_p, obj_id=obj_id_p)
        assert obj_id_p in self.storage
        assert obj_id_m not in self.storage

    def test_add_get_w_id(self):
        content, obj_id = self.hash_content(b"add_get_w_id")
        self.storage.add(content, obj_id=obj_id)
        self.assertContentMatch(obj_id, content)

    def test_add_get_w_composite_id(self):
        content, obj_id = self.compositehash_content(b"add_get_w_id")
        self.storage.add(content, obj_id=obj_id)
        self.assertContentMatch(obj_id, content)

    def test_add_twice(self):
        content, obj_id = self.hash_content(b"add_twice")
        self.storage.add(content, obj_id=obj_id)
        self.assertContentMatch(obj_id, content)
        self.storage.add(content, obj_id=obj_id, check_presence=False)
        self.assertContentMatch(obj_id, content)

    def test_add_big(self):
        content, obj_id = self.hash_content(b"add_big" * 1024 * 1024)
        self.storage.add(content, obj_id=obj_id)
        self.assertContentMatch(obj_id, content)

    def test_add_get_batch(self):
        content1, obj_id1 = self.hash_content(b"add_get_batch_1")
        content2, obj_id2 = self.hash_content(b"add_get_batch_2")
        self.storage.add(content1, obj_id1)
        self.storage.add(content2, obj_id2)
        cr1, cr2 = self.storage.get_batch([obj_id1, obj_id2])
        assert cr1 == content1
        assert cr2 == content2

    def test_add_get_batch_composite(self):
        content1, obj_id1 = self.compositehash_content(b"add_get_batch_1")
        content2, obj_id2 = self.compositehash_content(b"add_get_batch_2")
        self.storage.add(content1, obj_id1)
        self.storage.add(content2, obj_id2)
        cr1, cr2 = self.storage.get_batch([obj_id1, obj_id2])
        assert cr1 == content1
        assert cr2 == content2

    def test_get_batch_unexisting_content(self):
        content, obj_id = self.hash_content(b"get_batch_unexisting_content")
        result = list(self.storage.get_batch([obj_id]))
        assert len(result) == 1
        assert result[0] is None

    def test_restore_content(self):
        self.storage.allow_delete = True

        valid_content, valid_obj_id = self.hash_content(b"restore_content")
        invalid_content = b"unexpected content"
        self.storage.add(invalid_content, valid_obj_id)
        with pytest.raises((ObjCorruptedError, ObjNotFoundError)):
            # raise Corrupted except read only storage that raises NotFound,
            self.storage.check(valid_obj_id)
        assert valid_obj_id in self.storage
        self.storage.restore(valid_content, valid_obj_id)
        self.assertContentMatch(valid_obj_id, valid_content)

    def test_get_missing(self):
        content, obj_id = self.hash_content(b"get_missing")
        with pytest.raises(ObjNotFoundError) as e:
            self.storage.get(obj_id)

        assert obj_id in e.value.args

    def test_get_missing_composite(self):
        content, obj_id = self.compositehash_content(b"get_missing")
        with pytest.raises(ObjNotFoundError) as e:
            self.storage.get(obj_id)

        assert obj_id in e.value.args

    def test_check_missing(self):
        content, obj_id = self.hash_content(b"check_missing")
        with pytest.raises(ObjNotFoundError):
            self.storage.check(obj_id)

    def test_check_missing_composite(self):
        content, obj_id = self.compositehash_content(b"check_missing")
        with pytest.raises(ObjNotFoundError):
            self.storage.check(obj_id)

    def test_check_present(self):
        content, obj_id = self.hash_content(b"check_present")
        self.storage.add(content, obj_id)
        try:
            self.storage.check(obj_id)
        except ObjCorruptedError:
            self.fail("Integrity check failed")

    def test_check_present_composite(self):
        content, obj_id = self.compositehash_content(b"check_present")
        self.storage.add(content, obj_id)
        try:
            self.storage.check(obj_id)
        except ObjCorruptedError:
            self.fail("Integrity check failed")

    def test_delete_missing(self):
        self.storage.allow_delete = True
        content, obj_id = self.hash_content(b"missing_content_to_delete")
        with pytest.raises(ObjNotFoundError):
            self.storage.delete(obj_id)

    def test_delete_missing_composite(self):
        self.storage.allow_delete = True
        content, obj_id = self.compositehash_content(b"missing_content_to_delete")
        with pytest.raises(ObjNotFoundError):
            self.storage.delete(obj_id)

    def test_delete_present(self):
        self.storage.allow_delete = True
        content, obj_id = self.hash_content(b"content_to_delete")
        self.storage.add(content, obj_id=obj_id)
        assert self.storage.delete(obj_id)
        with pytest.raises(ObjNotFoundError):
            self.storage.get(obj_id)

    def test_delete_present_composite(self):
        self.storage.allow_delete = True
        content, obj_id = self.compositehash_content(b"content_to_delete")
        self.storage.add(content, obj_id=obj_id)
        assert self.storage.delete(obj_id)
        with pytest.raises(ObjNotFoundError):
            self.storage.get(obj_id)

    def test_delete_not_allowed(self):
        self.storage.allow_delete = False
        content, obj_id = self.hash_content(b"content_to_delete")
        self.storage.add(content, obj_id=obj_id)
        with pytest.raises(PermissionError):
            self.storage.delete(obj_id)

    def test_delete_not_allowed_by_default(self):
        content, obj_id = self.hash_content(b"content_to_delete")
        self.storage.add(content, obj_id=obj_id)
        with pytest.raises(PermissionError):
            self.storage.delete(obj_id)

    def test_add_batch(self):
        contents = {}
        expected_content_add = 0
        expected_content_add_bytes = 0
        for i in range(50):
            content = b"Test content %02d" % i
            content, obj_id = self.hash_content(content)
            contents[obj_id] = content
            expected_content_add_bytes += len(content)
            expected_content_add += 1

        ret = self.storage.add_batch(contents)

        assert ret == {
            "object:add": expected_content_add,
            "object:add:bytes": expected_content_add_bytes,
        }

        for obj_id in contents:
            assert obj_id in self.storage

    def test_add_batch_list(self):
        contents = []
        expected_content_add = 0
        expected_content_add_bytes = 0
        for i in range(50):
            content = b"Test content %02d" % i
            content, obj_id = self.hash_content(content)
            contents.append((obj_id, content))
            expected_content_add_bytes += len(content)
            expected_content_add += 1

        ret = self.storage.add_batch(contents)

        assert ret == {
            "object:add": expected_content_add,
            "object:add:bytes": expected_content_add_bytes,
        }

        for obj_id, content in contents:
            assert obj_id in self.storage

    def test_add_batch_list_composite(self):
        contents = []
        expected_content_add = 0
        expected_content_add_bytes = 0
        for i in range(50):
            content = b"Test content %02d" % i
            content, obj_id = self.compositehash_content(content)
            contents.append((obj_id, content))
            expected_content_add_bytes += len(content)
            expected_content_add += 1

        ret = self.storage.add_batch(contents)

        assert ret == {
            "object:add": expected_content_add,
            "object:add:bytes": expected_content_add_bytes,
        }

        for obj_id, content in contents:
            assert obj_id in self.storage

    def test_content_iterator(self):
        sto_obj_ids = list(iter(self.storage))
        assert not sto_obj_ids
        obj_ids = self.fill_objstorage(self.num_objects)

        sto_obj_ids = list(self.storage)
        assert len(sto_obj_ids) == len(obj_ids)
        assert sto_obj_ids == obj_ids

    @pytest.mark.skip_on_cloud
    def test_list_content_all(self):
        assert self.num_objects > 100
        all_ids = self.fill_objstorage(self.num_objects)

        ids = list(self.storage.list_content(limit=None))
        assert len(ids) == len(all_ids)
        assert ids == all_ids

    @pytest.mark.skip_on_cloud
    def test_list_content_limit(self):
        assert self.num_objects > 100
        all_ids = self.fill_objstorage(self.num_objects)

        ids = list(self.storage.list_content(limit=10))
        assert len(ids) == 10
        assert ids == all_ids[:10]

    @pytest.mark.skip_on_cloud
    def test_list_content_limit_and_last(self):
        assert self.num_objects > 110
        all_ids = self.fill_objstorage(self.num_objects)

        id0 = self.num_objects - 105
        ids = list(self.storage.list_content(last_obj_id=all_ids[id0], limit=100))
        assert len(ids) == 100
        assert ids == all_ids[id0 + 1 : id0 + 101]

        # check proper behavior at the end of the range
        id0 = self.num_objects - 51
        ids = list(self.storage.list_content(last_obj_id=all_ids[id0], limit=100))
        assert len(ids) == 50
        assert ids == all_ids[-50:]

        # check proper behavior after the end of the range
        ids = list(self.storage.list_content(last_obj_id=all_ids[-1], limit=100))
        assert not ids

        ids = list(
            self.storage.list_content(last_obj_id={"sha1": b"\xff" * 20}, limit=100)
        )
        assert not ids

    def test_download_url(self):
        content = b"foo"
        obj_id = compute_hash(content)
        self.storage.add(content, obj_id)
        url = self.storage.download_url(obj_id)
        if url is not None:
            decompress = decompressors[self.storage.compression]().decompress
            assert decompress(requests.get(url).content) == content

            with pytest.raises(ObjNotFoundError):
                self.storage.download_url(b"\xff" * 20)
